/*
 * Copyright (C) 2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/domain/batch_delete_domains.hh"

#include "libfred/registrable_object/domain/delete_domain.hh"
#include "libfred/registrable_object/domain/get_domain_state.hh"
#include "libfred/registrable_object/domain/info_domain.hh"
#include "libfred/registrar/info_registrar.hh"
#include "libfred/db_settings.hh"

#include <boost/optional.hpp>

#include <algorithm>


namespace Fred {
namespace Registry {
namespace Domain {

namespace {

template <typename F>
auto protect_by_savepoint(LibFred::OperationContext& ctx, F&& fnc)
{
    class Savepoint
    {
    public:
        Savepoint(const LibFred::OperationContext& ctx)
            : ctx_{ctx}
        {
            ctx_.get_conn().exec("SAVEPOINT _batch_delete_domains_savepoint_");
        }
        ~Savepoint()
        {
            try
            {
                ctx_.get_conn().exec("RELEASE SAVEPOINT _batch_delete_domains_savepoint_");
            }
            catch (...) { }
        }
        void rollback() const
        {
            try
            {
                ctx_.get_conn().exec("ROLLBACK TO SAVEPOINT _batch_delete_domains_savepoint_");
            }
            catch (...) { }
        }
    private:
        const LibFred::OperationContext& ctx_;
    } savepoint{ctx};
    try
    {
        return std::forward<F>(fnc)(ctx);
    }
    catch (...)
    {
        savepoint.rollback();
        throw;
    }
}

constexpr auto no_registrar_id = 0;

bool not_a_registrar(const LibFred::InfoRegistrarData& registrar_info)
{
    return registrar_info.id == no_registrar_id;
}

boost::optional<BatchDeleteDomainsReply::NotDeleted> try_to_delete(
        LibFred::OperationContext& ctx,
        const BatchDeleteDomainsRequest::Domain& domain,
        const LibFred::InfoRegistrarData& by_registrar)
{
    const auto not_deleted = [&](BatchDeleteDomainsReply::NotDeleted::Reason reason)
    {
        BatchDeleteDomainsReply::NotDeleted result{};
        result.domain_id = domain.domain_id;
        result.reason = reason;
        return result;
    };
    if (not_a_registrar(by_registrar))
    {
        return not_deleted(BatchDeleteDomainsReply::NotDeleted::Reason::insufficient_permissions);
    }
    try
    {
        const auto libfred_domain_uuid =
                ::Util::make_strong<LibFred::RegistrableObject::Domain::DomainUuid>(
                        get_raw_value_from(domain.domain_id));
        const auto domain_info = protect_by_savepoint(
                ctx,
                [&](LibFred::OperationContext& ctx)
                {
                    return LibFred::InfoDomainByUuid{libfred_domain_uuid}.exec(
                            LibFred::OperationContextLockingForUpdate{ctx}).info_domain_data;
                });
        const bool insufficient_permissions =
                (domain_info.sponsoring_registrar_handle != by_registrar.handle) &&
                (by_registrar.system.isnull() || !by_registrar.system.get_value());
        if (insufficient_permissions)
        {
            return not_deleted(BatchDeleteDomainsReply::NotDeleted::Reason::insufficient_permissions);
        }
        if (get_raw_value_from(domain.domain_history_id) != get_raw_value_from(domain_info.history_uuid))
        {
            return not_deleted(BatchDeleteDomainsReply::NotDeleted::Reason::domain_history_id_mismatch);
        }
        const auto domain_state = protect_by_savepoint(
                ctx,
                [&](LibFred::OperationContext& ctx)
                {
                    return LibFred::RegistrableObject::Domain::GetDomainStateByUuid{libfred_domain_uuid}
                            .exec(ctx);
                });
        if (domain_state.are_set_any_of<LibFred::RegistrableObject::Domain::ServerDeleteProhibited>())
        {
            return not_deleted(BatchDeleteDomainsReply::NotDeleted::Reason::domain_blocked);
        }
        protect_by_savepoint(
                ctx,
                [&](LibFred::OperationContext& ctx)
                {
                    LibFred::DeleteDomainById{domain_info.id}.exec(ctx);
                });
        return boost::none;
    }
    catch (const LibFred::InfoDomainByUuid::Exception& e)
    {
        if (e.is_set_unknown_domain_uuid())
        {
            return not_deleted(BatchDeleteDomainsReply::NotDeleted::Reason::domain_does_not_exist);
        }
        throw;
    }
}

} // namespace Fred::Registry::Domain::{anonymous}

BatchDeleteDomainsReply::Data batch_delete_domains(
        const BatchDeleteDomainsRequest& request,
        const boost::optional<std::string>& by_registrar)
{
    BatchDeleteDomainsReply::Data reply;
    LibFred::OperationContextCreator ctx;
    const auto registrar_info = [&]()
    {
        if (by_registrar != boost::none)
        {
            try
            {
                return protect_by_savepoint(
                        ctx,
                        [&](LibFred::OperationContext& ctx)
                        {
                            return LibFred::InfoRegistrarByHandle{*by_registrar}.exec(ctx).info_registrar_data;
                        });
            }
            catch (...) { }
        }
        LibFred::InfoRegistrarData no_registrar{};
        no_registrar.id = no_registrar_id;
        return no_registrar;
    }();
    std::for_each(begin(request.domains), end(request.domains), [&](auto&& domain)
            {
                const auto result = try_to_delete(ctx, domain, registrar_info);
                if (result != boost::none)
                {
                    reply.not_deleted_domains.push_back(*result);
                }
            });
    ctx.commit_transaction();
    return reply;
}

} // namespace Fred::Registry::Domain
} // namespace Fred::Registry
} // namespace Fred
