/*
 * Copyright (C) 2019-2022  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */
#include "src/domain/domain_info.hh"

#include "src/util/strong_type.hh"
#include "src/fred_unwrap.hh"

#include "libfred/registrable_object/domain/info_domain.hh"
#include "libfred/db_settings.hh"

namespace Fred {
namespace Registry {
namespace Domain {

DomainInfoReply::Data domain_info(const DomainInfoRequest& request)
{
    LibFred::OperationContextCreator ctx_provider;
    LibFred::OperationContextLockingForShare ctx(ctx_provider);
    try
    {
        if (request.domain_history_id != boost::none)
        {
            const auto domain_history_uuid = LibFred::RegistrableObject::make_history_uuid_of<LibFred::Object_Type::domain>(
                    get_raw_value_from(*request.domain_history_id));
            return fred_unwrap(LibFred::InfoDomainByHistoryUuid(domain_history_uuid).exec(ctx));
        }
        const auto domain_uuid = LibFred::RegistrableObject::make_uuid_of<LibFred::Object_Type::domain>(
                get_raw_value_from(request.domain_id));
        return fred_unwrap(LibFred::InfoDomainByUuid(domain_uuid).exec(ctx));
    }
    catch (const LibFred::InfoDomainByUuid::Exception& e)
    {
        if (e.is_set_unknown_domain_uuid())
        {
            throw DomainInfoReply::Exception::DomainDoesNotExist(request);
        }
        throw InternalServerError("unexpected problem signalized by LibFred::DomainInfoByUuid::Exception");;
    }
    catch (const LibFred::InfoDomainByHistoryUuid::Exception& e)
    {
        if (e.is_set_unknown_domain_history_uuid())
        {
            throw DomainInfoReply::Exception::DomainDoesNotExist(request);
        }
        throw InternalServerError("unexpected problem signalized by LibFred::InfoDomainByHistoryUuid::Exception");;
    }
    catch (const std::exception& e)
    {
        throw InternalServerError(std::string("expected unknown exception: ") + e.what());
    }
    catch (...)
    {
        throw InternalServerError("unexpected unknown exception");
    }
}

DomainIdReply::Data get_domain_id(const DomainIdRequest& request)
{
    LibFred::OperationContextCreator ctx;
    try
    {
        const auto db_res = ctx.get_conn().exec_params(
                "SELECT obr.uuid, h.uuid "
                "FROM object_registry obr "
                "JOIN history h ON h.id = obr.historyid "
                "WHERE obr.erdate IS NULL AND "
                      "obr.type = get_object_type_id('domain') AND "
                      "obr.name = LOWER($1::TEXT) "
                "FOR SHARE OF obr",
                Database::QueryParams{get_raw_value_from(request.fqdn)});
        if (db_res.size() == 0)
        {
            throw DomainIdReply::Exception::DomainDoesNotExist(request);
        }
        if (1 < db_res.size())
        {
            throw InternalServerError("too many domains selected by fqdn");
        }
        auto result = DomainIdReply::Data{};
        result.domain_id = Util::make_strong<DomainId>(
                boost::uuids::string_generator{}(static_cast<std::string>(db_res[0][0])));
        result.domain_history_id = Util::make_strong<DomainHistoryId>(
                boost::uuids::string_generator{}(static_cast<std::string>(db_res[0][1])));
        return result;
    }
    catch (const DomainIdReply::Exception::DomainDoesNotExist&)
    {
        throw;
    }
    catch (const InternalServerError&)
    {
        throw;
    }
    catch (const std::exception& e)
    {
        throw InternalServerError(std::string("expected unknown exception: ") + e.what());
    }
    catch (...)
    {
        throw InternalServerError("unexpected unknown exception");
    }
}

} // namespace Fred::Registry::Domain
} // namespace Fred::Registry
} // namespace Fred
