/*
 * Copyright (C) 2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/domain/get_domains_by_holder.hh"

#include "libfred/registrable_object/contact/info_contact.hh"
#include "libfred/registrable_object/domain/info_domain.hh"
#include "libfred/db_settings.hh"

#include <algorithm>

namespace Fred {
namespace Registry {
namespace Domain {

GetDomainsByHolderReply::Data get_domains_by_holder(const GetDomainsByHolderRequest& request)
{
    GetDomainsByHolderReply::Data reply;
    const auto domains = [](const GetDomainsByHolderRequest& request, GetDomainsByHolderReply::Data& reply)
    {
        LibFred::OperationContextCreator ctx;
        const auto holder_info = [&]()
        {
            const auto libfred_contact_uuid =
                    ::Util::make_strong<LibFred::RegistrableObject::Contact::ContactUuid>(
                            get_raw_value_from(request.domain_holder_id));
            try
            {
                return LibFred::InfoContactByUuid{libfred_contact_uuid}.exec(
                        LibFred::OperationContextLockingForShare{ctx}).info_contact_data;
            }
            catch (const LibFred::InfoContactByUuid::Exception& e)
            {
                if (e.is_set_unknown_contact_uuid())
                {
                    throw GetDomainsByHolderReply::Exception::DomainHolderDoesNotExist{request};
                }
                throw;
            }
        }();
        if (request.domain_holder_history_id != boost::none)
        {
            if (get_raw_value_from(holder_info.history_uuid) != get_raw_value_from(*request.domain_holder_history_id))
            {
                throw GetDomainsByHolderReply::Exception::DomainHolderDoesNotExist{request};
            }
        }
        reply.domain_holder_id = Util::make_strong<Contact::ContactId>(
                get_raw_value_from(holder_info.uuid));
        reply.domain_holder_history_id = Util::make_strong<Contact::ContactHistoryId>(
                get_raw_value_from(holder_info.history_uuid));
        auto domains = LibFred::InfoDomainByRegistrantHandle{holder_info.handle}.exec(ctx);
        ctx.commit_transaction();
        return domains;
    }(request, reply);
    std::for_each(begin(domains), end(domains), [&](auto&& info_domain_result)
            {
                const auto& domain_data = info_domain_result.info_domain_data;
                auto domain = GetDomainsByHolderReply::Data::Domain{};
                domain.domain_id = Util::make_strong<DomainId>(
                        get_raw_value_from(domain_data.uuid));
                domain.domain_history_id = Util::make_strong<DomainHistoryId>(
                        get_raw_value_from(domain_data.history_uuid));
                domain.fqdn = Util::make_strong<Fqdn>(domain_data.fqdn);
                reply.domains.push_back(std::move(domain));
            });
    return reply;
}

} // namespace Fred::Registry::Domain
} // namespace Fred::Registry
} // namespace Fred
