/*
 * Copyright (C) 2018-2022  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/cfg.hh"

#include "src/contact/contact_common_types.hh"
#include "src/contact/contact_data_history.hh"
#include "src/contact/contact_handle_history.hh"
#include "src/contact/contact_info.hh"
#include "src/contact/contact_state.hh"
#include "src/contact/contact_state_flags.hh"
#include "src/contact/contact_state_history.hh"
#include "src/contact/update_contact.hh"

#include "src/diagnostics/about.hh"
#include "src/diagnostics/status.hh"

#include "src/domain/batch_delete_domains.hh"
#include "src/domain/domain_common_types.hh"
#include "src/domain/domain_data_history.hh"
#include "src/domain/domain_info.hh"
#include "src/domain/domain_info_reply.hh"
#include "src/domain/domain_info_request.hh"
#include "src/domain/domain_state.hh"
#include "src/domain/domain_state_flags.hh"
#include "src/domain/domain_state_history.hh"
#include "src/domain/fqdn_history.hh"
#include "src/domain/get_domains_by_contact.hh"
#include "src/domain/get_domains_by_holder.hh"
#include "src/domain/get_domains_notify_info.hh"
#include "src/domain/manage_domain_state_flags.hh"
#include "src/domain/update_domains_additional_notify_info.hh"
#include "src/domain/update_domains_additional_notify_info_reply.hh"

#include "src/keyset/keyset_common_types.hh"
#include "src/keyset/keyset_data_history.hh"
#include "src/keyset/keyset_handle_history.hh"
#include "src/keyset/keyset_info.hh"
#include "src/keyset/keyset_info_reply.hh"
#include "src/keyset/keyset_info_request.hh"
#include "src/keyset/keyset_state.hh"
#include "src/keyset/keyset_state_flags.hh"
#include "src/keyset/keyset_state_history.hh"

#include "src/nsset/nsset_common_types.hh"
#include "src/nsset/nsset_data_history.hh"
#include "src/nsset/nsset_handle_history.hh"
#include "src/nsset/nsset_info.hh"
#include "src/nsset/nsset_info_reply.hh"
#include "src/nsset/nsset_info_request.hh"
#include "src/nsset/nsset_state.hh"
#include "src/nsset/nsset_state_flags.hh"
#include "src/nsset/nsset_state_history.hh"

#include "src/contact/search_contact.hh"
#include "src/domain/search_domain.hh"
#include "src/keyset/search_keyset.hh"
#include "src/nsset/search_nsset.hh"

#include "src/registrar/registrar_credit.hh"
#include "src/registrar/registrar_info.hh"
#include "src/registrar/registrar_list.hh"

#include "src/proto_unwrap.hh"
#include "src/proto_wrap.hh"

#include "util/log/add_log_device.hh"
#include "util/log/context.hh"
#include "util/log/logger.hh"
#include "util/random/random.hh"

#include "libfred/db_settings.hh"

#include "libfred/registrable_object/contact/contact_state.hh"
#include "libfred/registrable_object/domain/domain_state.hh"
#include "libfred/registrable_object/keyset/keyset_state.hh"
#include "libfred/registrable_object/nsset/nsset_state.hh"

#include "fred_api/registry/contact/service_contact_grpc.pb.h"
#include "fred_api/registry/contact/service_contact_grpc.grpc.pb.h"
#include "fred_api/registry/contact/service_search_contact_grpc.pb.h"
#include "fred_api/registry/contact/service_search_contact_grpc.grpc.pb.h"
#include "fred_api/registry/domain/service_admin_grpc.pb.h"
#include "fred_api/registry/domain/service_admin_grpc.grpc.pb.h"
#include "fred_api/registry/domain/service_domain_grpc.pb.h"
#include "fred_api/registry/domain/service_domain_grpc.grpc.pb.h"
#include "fred_api/registry/domain/service_search_domain_grpc.pb.h"
#include "fred_api/registry/domain/service_search_domain_grpc.grpc.pb.h"
#include "fred_api/registry/keyset/service_keyset_grpc.pb.h"
#include "fred_api/registry/keyset/service_keyset_grpc.grpc.pb.h"
#include "fred_api/registry/keyset/service_search_keyset_grpc.pb.h"
#include "fred_api/registry/keyset/service_search_keyset_grpc.grpc.pb.h"
#include "fred_api/registry/nsset/service_nsset_grpc.pb.h"
#include "fred_api/registry/nsset/service_nsset_grpc.grpc.pb.h"
#include "fred_api/registry/nsset/service_search_nsset_grpc.pb.h"
#include "fred_api/registry/nsset/service_search_nsset_grpc.grpc.pb.h"

#include "fred_api/registry/registrar/service_registrar_grpc.pb.h"
#include "fred_api/registry/registrar/service_registrar_grpc.grpc.pb.h"

#include <grpc++/grpc++.h>

#include <boost/algorithm/string.hpp>
#include <boost/date_time.hpp>
#include <boost/format.hpp>
#include <boost/optional.hpp>
#include <boost/program_options.hpp>

#include <pthread.h>

#include <csignal>
#include <cstring>
#include <cstdlib>

#include <chrono>
#include <iostream>
#include <memory>
#include <string>

namespace Fred {
namespace Registry {

namespace {

std::string create_ctx_name(const std::string& _name)
{
    return str(boost::format("%1%-<%2%>") % _name % Random::Generator().get(0, 10000));
}

std::string create_ctx_function_name(const char* fnc)
{
    std::string name(fnc);
    std::replace(name.begin(), name.end(), '_', '-');
    return name;
}

class LogContext
{
public:
    LogContext(const std::string& _op_name)
        : ctx_server_(create_ctx_name("registry")),
          ctx_operation_(_op_name)
    { }
private:
    Logging::Context ctx_server_;
    Logging::Context ctx_operation_;
};

#define LOGGING_CONTEXT(CTX_VAR) const LogContext CTX_VAR(create_ctx_function_name(__FUNCTION__))

grpc::Status on_internal_server_error(const std::exception& e)
{
    LOGGER.error(std::string{"InternalServerError caught: "} + e.what());
    return grpc::Status(grpc::StatusCode::UNKNOWN, "InternalServerError");
}

grpc::Status on_invalid_argument(const std::exception& e)
{
    LOGGER.warning(std::string{"InvalidArgument caught: "} + e.what());
    return grpc::Status(grpc::StatusCode::INVALID_ARGUMENT, "InvalidArgument");
}

grpc::Status on_std_exception(const std::exception& e)
{
    LOGGER.error(std::string{"std::exception caught: "} + e.what());
    return grpc::Status(grpc::StatusCode::UNKNOWN, "Unknown exception based on std::exception");
}

grpc::Status on_unknown_exception()
{
    LOGGER.error("unknown exception caught");
    return grpc::Status(grpc::StatusCode::UNKNOWN, "Unknown exception");
}

}//namespace Fred::Registry::{anonymous}

namespace Contact {

namespace {

class ContactService final : public Api::Contact::Contact::Service
{
private:
    grpc::Status get_contact_handle_history(
            grpc::ServerContext*,
            const Api::Contact::ContactHandleHistoryRequest* _api_request,
            Api::Contact::ContactHandleHistoryReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_contact_handle_history requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: contact_handle_history(" + request.to_string() + ")");
            const auto handle_history = contact_handle_history(request);
            proto_wrap(handle_history, _api_response);
            LOGGER.debug("server: answer = " + handle_history.to_string());
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_contact_history(
            grpc::ServerContext*,
            const Api::Contact::ContactHistoryRequest* _api_request,
            Api::Contact::ContactHistoryReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_contact_history requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: contact_history(" + request.to_string() + ")");
            const auto data_history = contact_data_history(request);
            proto_wrap(data_history, _api_response);
            LOGGER.debug("server: answer = " + data_history.to_string());
            return grpc::Status::OK;
        }
        catch (const ContactDataHistoryReply::Exception::ContactDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const ContactDataHistoryReply::Exception::InvalidHistoryInterval& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_contact_info(
            grpc::ServerContext*,
            const Api::Contact::ContactInfoRequest* _api_request,
            Api::Contact::ContactInfoReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_contact_info requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: contact_info(" + request.to_string() + ")");
            const auto info = contact_info(request);
            proto_wrap(info, _api_response);
            LOGGER.debug("server: answer = " + info.to_string());
            return grpc::Status::OK;
        }
        catch (const ContactInfoReply::Exception::ContactDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_contact_id(
            grpc::ServerContext*,
            const Api::Contact::ContactIdRequest* _api_request,
            Api::Contact::ContactIdReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_contact_id requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: get_contact_id(" + request.to_string() + ")");
            const auto id = Fred::Registry::Contact::get_contact_id(request);
            proto_wrap(id, _api_response);
            LOGGER.debug("server: answer = " + id.to_string());
            return grpc::Status::OK;
        }
        catch (const ContactIdReply::Exception::ContactDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_contact_state(
            grpc::ServerContext*,
            const Api::Contact::ContactStateRequest* _api_request,
            Api::Contact::ContactStateReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_contact_state requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: contact_state(" + request.to_string() + ")");
            const auto state = contact_state(request);
            proto_wrap(state, _api_response);
            LOGGER.debug("server: answer = " + state.to_string());
            return grpc::Status::OK;
        }
        catch (const ContactStateReply::Exception::ContactDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_contact_state_history(
            grpc::ServerContext*,
            const Api::Contact::ContactStateHistoryRequest* _api_request,
            Api::Contact::ContactStateHistoryReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_contact_state_history requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: contact_state_history(" + request.to_string() + ")");
            const auto state_history = contact_state_history(request);
            proto_wrap(state_history, _api_response);
            LOGGER.debug("server: answer = " + state_history.to_string());
            return grpc::Status::OK;
        }
        catch (const ContactStateHistoryReply::Exception::ContactDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const ContactStateHistoryReply::Exception::InvalidHistoryInterval& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_contact_state_flags(
            grpc::ServerContext*,
            const google::protobuf::Empty*,
            Api::Contact::ContactStateFlagsReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_contact_state_flags requested");
            const auto info = get_contact_state_flags_info();
            proto_wrap(info, _api_response);
            LOGGER.debug("server: answer = " + to_string(info));
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status update_contact(
            grpc::ServerContext*,
            const Api::Contact::UpdateContactRequest* _api_request,
            Api::Contact::UpdateContactReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: update_contact requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: update_contact(" + request.to_string() + ")");
            const auto result = Contact::update_contact(request);
            proto_wrap(result, _api_response);
            LOGGER.debug("server: answer = " + result.to_string());
            return grpc::Status::OK;
        }
        catch (const UpdateContactReply::Exception::ContactDoesNotExist& e)
        {
            LOGGER.info(std::string{"server: "} + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const UpdateContactReply::Exception::NotCurrentVersion& e)
        {
            LOGGER.info(std::string{"server: "} + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const UpdateContactReply::Exception::InvalidArgument& e)
        {
            return on_invalid_argument(static_cast<const std::exception&>(e));
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
};

class SearchContactService final : public Api::Contact::SearchContact::SearchContact::Service
{
private:
    grpc::Status search_contact(
            grpc::ServerContext*,
            const Api::Contact::SearchContactRequest* _api_request,
            Api::Contact::SearchContactReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: search_contact requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: search_contact(" + request.to_string() + ")");
            const auto response = Fred::Registry::Contact::search_contact(request);
            proto_wrap(response, _api_response);
            LOGGER.debug("server: answer = " + response.to_string());
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status search_contact_history(
        grpc::ServerContext*,
        const Api::Contact::SearchContactHistoryRequest* _api_request,
        Api::Contact::SearchContactHistoryReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: search_contact_history requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: search_contact_history(" + request.to_string() + ")");
            const auto response = Fred::Registry::Contact::search_contact(request);
            proto_wrap(response, _api_response);
            LOGGER.debug("server: answer = " + response.to_string());
            return grpc::Status::OK;
        }
        catch (const SearchContactHistoryReply::Exception::ChronologyViolation& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
};

}//namespace Fred::Registry::Contact::{anonymous}

}//namespace Fred::Registry::Contact

namespace Domain {

namespace {

class AdminService final : public Api::Domain::Admin::Service
{
private:
    grpc::Status batch_delete_domains(
            grpc::ServerContext*,
            grpc::ServerReaderWriter<Api::Domain::BatchDeleteDomainsReply, ::Fred::Registry::Api::Domain::BatchDeleteDomainsRequest>* api_stream) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: batch_delete_domains requested");
            ::Fred::Registry::Api::Domain::BatchDeleteDomainsRequest request_chunk;
            Domain::BatchDeleteDomainsRequest request{};
            while (api_stream->Read(&request_chunk))
            {
                ::Fred::Registry::proto_unwrap_back(request_chunk, request);
            }
            const auto reply = ::Fred::Registry::Domain::batch_delete_domains(
                    request,
                    Cfg::Options::get().fred.registrar_originator);
            std::size_t offset = 0;
            while (true)
            {
                Api::Domain::BatchDeleteDomainsReply response{};
                offset = proto_wrap(reply, offset, response.mutable_data());
                const bool last_chunk = offset == 0;
                if (last_chunk)
                {
                    api_stream->WriteLast(response, ::grpc::WriteOptions{});
                    break;
                }
                api_stream->Write(response);
            }
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status manage_domain_state_flags(
            grpc::ServerContext*,
            grpc::ServerReaderWriter<Api::Domain::ManageDomainStateFlagsReply, Api::Domain::ManageDomainStateFlagsRequest>* api_stream) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: manage_domain_state_flags requested");
            ::Fred::Registry::Api::Domain::ManageDomainStateFlagsRequest request_chunk;
            Domain::ManageDomainStateFlagsRequest request{};
            while (api_stream->Read(&request_chunk))
            {
                ::Fred::Registry::proto_unwrap_back(request_chunk, request);
            }
            const auto reply = ::Fred::Registry::Domain::manage_domain_state_flags(
                    request,
                    Cfg::Options::get().fred.registrar_originator);
            std::size_t offset = 0;
            while (true)
            {
                Api::Domain::ManageDomainStateFlagsReply response{};
                offset = proto_wrap(reply, offset, response.mutable_data());
                const bool last_chunk = offset == 0;
                if (last_chunk)
                {
                    api_stream->WriteLast(response, ::grpc::WriteOptions{});
                    break;
                }
                api_stream->Write(response);
            }
            return grpc::Status::OK;
        }
        catch (const Domain::ManageDomainStateFlagsReply::Exception::DomainDoesNotExist& e)
        {
            LOGGER.info(std::string{"server: "} + e.what());
            Api::Domain::ManageDomainStateFlagsReply api_response;
            proto_wrap(e, &api_response);
            api_stream->Write(api_response, ::grpc::WriteOptions{}.set_last_message());
            return grpc::Status::OK;
        }
        catch (const Domain::ManageDomainStateFlagsReply::Exception::DomainStateFlagRequestDoesNotExist& e)
        {
            LOGGER.info(std::string{"server: "} + e.what());
            Api::Domain::ManageDomainStateFlagsReply api_response;
            proto_wrap(e, &api_response);
            api_stream->Write(api_response, ::grpc::WriteOptions{}.set_last_message());
            return grpc::Status::OK;
        }
        catch (const Domain::ManageDomainStateFlagsReply::Exception::DomainStateFlagDoesNotExist& e)
        {
            LOGGER.info(std::string{"server: "} + e.what());
            Api::Domain::ManageDomainStateFlagsReply api_response;
            proto_wrap(e, &api_response);
            api_stream->Write(api_response, ::grpc::WriteOptions{}.set_last_message());
            return grpc::Status::OK;
        }
        catch (const Domain::ManageDomainStateFlagsReply::Exception::DomainHistoryIdMismatch& e)
        {
            LOGGER.info(std::string{"server: "} + e.what());
            Api::Domain::ManageDomainStateFlagsReply api_response;
            proto_wrap(e, &api_response);
            api_stream->Write(api_response, ::grpc::WriteOptions{}.set_last_message());
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_domains_outzone_notify_info(
            grpc::ServerContext*,
            const Api::Domain::GetDomainsOutzoneNotifyInfoRequest* api_request,
            grpc::ServerWriter<Api::Domain::GetDomainsNotifyInfoReply>* api_response_writer) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_domains_outzone_notify_info requested");
            const auto request = proto_unwrap(*api_request);
            LOGGER.debug("server: get_domains_outzone_notify_info(" + request.to_string() + ")");
            const auto reply = Domain::get_domains_notify_info(request);
            std::size_t offset = 0;
            while (true)
            {
                Api::Domain::GetDomainsNotifyInfoReply response;
                offset = proto_wrap(reply, offset, response.mutable_data());
                const bool last_chunk = offset == 0;
                if (last_chunk)
                {
                    api_response_writer->WriteLast(response, ::grpc::WriteOptions{});
                    break;
                }
                api_response_writer->Write(response);
            }
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status update_domains_outzone_additional_notify_info(
            ::grpc::ServerContext*,
            ::grpc::ServerReader<::Fred::Registry::Api::Domain::UpdateDomainsOutzoneAdditionalNotifyInfoRequest>* api_stream_reader,
            ::Fred::Registry::Api::Domain::UpdateDomainsAdditionalNotifyInfoReply* api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: update_domains_outzone_additional_notify_info requested");
            ::Fred::Registry::Api::Domain::UpdateDomainsOutzoneAdditionalNotifyInfoRequest request_chunk;
            Domain::UpdateDomainsAdditionalNotifyInfoRequest request{};
            while (api_stream_reader->Read(&request_chunk))
            {
                ::Fred::Registry::proto_unwrap_back(request_chunk, request);
            }
            ::Fred::Registry::Domain::update_domains_additional_notify_info(request);
            api_response->clear_exception();
            return grpc::Status::OK;
        }
        catch (const Domain::UpdateDomainsAdditionalNotifyInfoReply::Exception::DomainDoesNotExist& e)
        {
            LOGGER.info(std::string{"server: "} + e.what());
            proto_wrap(e, api_response);
            return grpc::Status::OK;
        }
        catch (const Domain::UpdateDomainsAdditionalNotifyInfoReply::Exception::InvalidDomainContactInfo& e)
        {
            LOGGER.info(std::string{"server: "} + e.what());
            proto_wrap(e, api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_domains_delete_notify_info(
            grpc::ServerContext*,
            const Api::Domain::GetDomainsDeleteNotifyInfoRequest* api_request,
            grpc::ServerWriter<Api::Domain::GetDomainsNotifyInfoReply>* api_response_writer) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_domains_delete_notify_info requested");
            const auto request = proto_unwrap(*api_request);
            LOGGER.debug("server: get_domains_delete_notify_info(" + request.to_string() + ")");
            const auto reply = Domain::get_domains_notify_info(request);
            std::size_t offset = 0;
            while (true)
            {
                Api::Domain::GetDomainsNotifyInfoReply response;
                offset = proto_wrap(reply, offset, response.mutable_data());
                const bool last_chunk = offset == 0;
                if (last_chunk)
                {
                    api_response_writer->WriteLast(response, ::grpc::WriteOptions{});
                    break;
                }
                api_response_writer->Write(response);
            }
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status update_domains_delete_additional_notify_info(
            ::grpc::ServerContext*,
            ::grpc::ServerReader<::Fred::Registry::Api::Domain::UpdateDomainsDeleteAdditionalNotifyInfoRequest>* api_stream_reader,
            ::Fred::Registry::Api::Domain::UpdateDomainsAdditionalNotifyInfoReply* api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: update_domains_delete_additional_notify_info requested");
            ::Fred::Registry::Api::Domain::UpdateDomainsDeleteAdditionalNotifyInfoRequest request_chunk;
            Domain::UpdateDomainsAdditionalNotifyInfoRequest request{};
            while (api_stream_reader->Read(&request_chunk))
            {
                ::Fred::Registry::proto_unwrap_back(request_chunk, request);
            }
            ::Fred::Registry::Domain::update_domains_additional_notify_info(request);
            api_response->clear_exception();
            return grpc::Status::OK;
        }
        catch (const Domain::UpdateDomainsAdditionalNotifyInfoReply::Exception::DomainDoesNotExist& e)
        {
            LOGGER.info(std::string{"server: "} + e.what());
            proto_wrap(e, api_response);
            return grpc::Status::OK;
        }
        catch (const Domain::UpdateDomainsAdditionalNotifyInfoReply::Exception::InvalidDomainContactInfo& e)
        {
            LOGGER.info(std::string{"server: "} + e.what());
            proto_wrap(e, api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
};

class DomainService final : public Api::Domain::Domain::Service
{
private:
    grpc::Status get_fqdn_history(
            grpc::ServerContext*,
            const Api::Domain::FqdnHistoryRequest* _api_request,
            Api::Domain::FqdnHistoryReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_fqdn_history requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: fqdn_history(" + request.to_string() + ")");
            const auto handle_history = fqdn_history(request);
            proto_wrap(handle_history, _api_response);
            LOGGER.debug("server: answer = " + handle_history.to_string());
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_domain_history(
            grpc::ServerContext*,
            const Api::Domain::DomainHistoryRequest* _api_request,
            Api::Domain::DomainHistoryReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_domain_history requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: domain_history(" + request.to_string() + ")");
            const auto data_history = domain_data_history(request);
            proto_wrap(data_history, _api_response);
            LOGGER.debug("server: answer = " + data_history.to_string());
            return grpc::Status::OK;
        }
        catch (const DomainDataHistoryReply::Exception::DomainDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const DomainDataHistoryReply::Exception::InvalidHistoryInterval& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_domain_info(
            grpc::ServerContext*,
            const Api::Domain::DomainInfoRequest* _api_request,
            Api::Domain::DomainInfoReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_domain_info requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: domain_info(" + request.to_string() + ")");
            const auto info = domain_info(request);
            proto_wrap(info, _api_response);
            LOGGER.debug("server: answer = " + info.to_string());
            return grpc::Status::OK;
        }
        catch (const DomainInfoReply::Exception::DomainDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_domain_id(
            grpc::ServerContext*,
            const Api::Domain::DomainIdRequest* _api_request,
            Api::Domain::DomainIdReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_domain_id requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: domain_id(" + request.to_string() + ")");
            const auto id = Fred::Registry::Domain::get_domain_id(request);
            proto_wrap(id, _api_response);
            LOGGER.debug("server: answer = " + id.to_string());
            return grpc::Status::OK;
        }
        catch (const DomainIdReply::Exception::DomainDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_domain_state(
            grpc::ServerContext*,
            const Api::Domain::DomainStateRequest* _api_request,
            Api::Domain::DomainStateReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_domain_state requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: domain_state(" + request.to_string() + ")");
            const auto state = domain_state(request);
            proto_wrap(state, _api_response);
            LOGGER.debug("server: answer = " + state.to_string());
            return grpc::Status::OK;
        }
        catch (const DomainStateReply::Exception::DomainDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_domain_state_history(
            grpc::ServerContext*,
            const Api::Domain::DomainStateHistoryRequest* _api_request,
            Api::Domain::DomainStateHistoryReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_domain_state_history requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: domain_state_history(" + request.to_string() + ")");
            const auto state_history = domain_state_history(request);
            proto_wrap(state_history, _api_response);
            LOGGER.debug("server: answer = " + state_history.to_string());
            return grpc::Status::OK;
        }
        catch (const DomainStateHistoryReply::Exception::DomainDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const DomainStateHistoryReply::Exception::InvalidHistoryInterval& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_domain_state_flags(
            grpc::ServerContext*,
            const google::protobuf::Empty*,
            Api::Domain::DomainStateFlagsReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_domain_state_flags requested");
            const auto info = get_domain_state_flags_info();
            proto_wrap(info, _api_response);
            LOGGER.debug("server: answer = " + to_string(info));
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_domains_by_contact(
            grpc::ServerContext*,
            const Api::Domain::DomainsByContactRequest* api_request,
            grpc::ServerWriter<Api::Domain::DomainsByContactReply>* api_response_writer) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_domains_by_contact requested");
            const auto request = proto_unwrap(*api_request);
            LOGGER.debug("server: get_domains_by_contact(" + request.to_string() + ")");
            const auto reply = Domain::get_domains_by_contact(request);
            std::size_t offset = 0;
            while (true)
            {
                Api::Domain::DomainsByContactReply response;
                offset = proto_wrap(reply, offset, response.mutable_data());
                const bool last_chunk = offset == 0;
                if (last_chunk)
                {
                    api_response_writer->WriteLast(response, ::grpc::WriteOptions{});
                    break;
                }
                api_response_writer->Write(response);
            }
            return grpc::Status::OK;
        }
        catch (const Fred::Registry::Domain::GetDomainsByContactReply::Exception::ContactDoesNotExist& e)
        {
            LOGGER.info(std::string{"server: "} + e.what());
            Fred::Registry::Api::Domain::DomainsByContactReply reply;
            proto_wrap(e, &reply);
            api_response_writer->WriteLast(reply, ::grpc::WriteOptions{});
            return grpc::Status::OK;
        }
        catch (const Fred::Registry::Domain::GetDomainsByContactReply::Exception::InvalidOrderBy& e)
        {
            LOGGER.info(std::string{"server: "} + e.what());
            Fred::Registry::Api::Domain::DomainsByContactReply reply;
            proto_wrap(e, &reply);
            api_response_writer->WriteLast(reply, ::grpc::WriteOptions{});
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_domains_by_holder(
            grpc::ServerContext*,
            const Api::Domain::DomainsByHolderRequest* api_request,
            grpc::ServerWriter<Api::Domain::DomainsByHolderReply>* api_response_writer) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_domains_by_holder requested");
            const auto request = proto_unwrap(*api_request);
            LOGGER.debug("server: get_domains_by_holder(" + request.to_string() + ")");
            const auto reply = Domain::get_domains_by_holder(request);
            std::size_t offset = 0;
            while (true)
            {
                Api::Domain::DomainsByHolderReply response;
                offset = proto_wrap(reply, offset, response.mutable_data());
                const bool last_chunk = offset == 0;
                if (last_chunk)
                {
                    api_response_writer->WriteLast(response, ::grpc::WriteOptions{});
                    break;
                }
                api_response_writer->Write(response);
            }
            return grpc::Status::OK;
        }
        catch (const Fred::Registry::Domain::GetDomainsByHolderReply::Exception::DomainHolderDoesNotExist& e)
        {
            LOGGER.info(std::string{"server: "} + e.what());
            Fred::Registry::Api::Domain::DomainsByHolderReply reply;
            proto_wrap(e, &reply);
            api_response_writer->WriteLast(reply, ::grpc::WriteOptions{});
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
};

class SearchDomainService final : public Api::Domain::SearchDomain::SearchDomain::Service
{
private:
    grpc::Status search_domain(
            grpc::ServerContext*,
            const Api::Domain::SearchDomainRequest* _api_request,
            Api::Domain::SearchDomainReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: search_domain requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: search_domain(" + request.to_string() + ")");
            const auto response = Fred::Registry::Domain::search_domain(request);
            proto_wrap(response, _api_response);
            LOGGER.debug("server: answer = " + response.to_string());
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status search_domain_history(
            grpc::ServerContext*,
            const Api::Domain::SearchDomainHistoryRequest* _api_request,
            Api::Domain::SearchDomainHistoryReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: search_domain_history requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: search_domain_history(" + request.to_string() + ")");
            const auto response = Fred::Registry::Domain::search_domain(request);
            proto_wrap(response, _api_response);
            LOGGER.debug("server: answer = " + response.to_string());
            return grpc::Status::OK;
        }
        catch (const SearchDomainHistoryReply::Exception::ChronologyViolation& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
};

}//namespace Fred::Registry::Domain::{anonymous}

}//namespace Fred::Registry::Domain

namespace Keyset {

namespace {

class KeysetService final : public Api::Keyset::Keyset::Service
{
private:
    grpc::Status get_keyset_handle_history(
            grpc::ServerContext*,
            const Api::Keyset::KeysetHandleHistoryRequest* _api_request,
            Api::Keyset::KeysetHandleHistoryReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_keyset_handle_history requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: keyset_handle_history(" + request.to_string() + ")");
            const auto handle_history = keyset_handle_history(request);
            proto_wrap(handle_history, _api_response);
            LOGGER.debug("server: answer = " + handle_history.to_string());
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_keyset_history(
            grpc::ServerContext*,
            const Api::Keyset::KeysetHistoryRequest* _api_request,
            Api::Keyset::KeysetHistoryReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_keyset_history requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: keyset_history(" + request.to_string() + ")");
            const auto data_history = keyset_data_history(request);
            proto_wrap(data_history, _api_response);
            LOGGER.debug("server: answer = " + data_history.to_string());
            return grpc::Status::OK;
        }
        catch (const KeysetDataHistoryReply::Exception::KeysetDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const KeysetDataHistoryReply::Exception::InvalidHistoryInterval& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_keyset_info(
            grpc::ServerContext*,
            const Api::Keyset::KeysetInfoRequest* _api_request,
            Api::Keyset::KeysetInfoReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_keyset_info requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: keyset_info(" + request.to_string() + ")");
            const auto info = keyset_info(request);
            proto_wrap(info, _api_response);
            LOGGER.debug("server: answer = " + info.to_string());
            return grpc::Status::OK;
        }
        catch (const KeysetInfoReply::Exception::KeysetDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_keyset_id(
            grpc::ServerContext*,
            const Api::Keyset::KeysetIdRequest* _api_request,
            Api::Keyset::KeysetIdReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_keyset_id requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: keyset_id(" + request.to_string() + ")");
            const auto id = Fred::Registry::Keyset::get_keyset_id(request);
            proto_wrap(id, _api_response);
            LOGGER.debug("server: answer = " + id.to_string());
            return grpc::Status::OK;
        }
        catch (const KeysetIdReply::Exception::KeysetDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_keyset_state(
            grpc::ServerContext*,
            const Api::Keyset::KeysetStateRequest* _api_request,
            Api::Keyset::KeysetStateReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_keyset_state requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: keyset_state(" + request.to_string() + ")");
            const auto state = keyset_state(request);
            proto_wrap(state, _api_response);
            LOGGER.debug("server: answer = " + state.to_string());
            return grpc::Status::OK;
        }
        catch (const KeysetStateReply::Exception::KeysetDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_keyset_state_history(
            grpc::ServerContext*,
            const Api::Keyset::KeysetStateHistoryRequest* _api_request,
            Api::Keyset::KeysetStateHistoryReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_keyset_state_history requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: keyset_state_history(" + request.to_string() + ")");
            const auto state_history = keyset_state_history(request);
            proto_wrap(state_history, _api_response);
            LOGGER.debug("server: answer = " + state_history.to_string());
            return grpc::Status::OK;
        }
        catch (const KeysetStateHistoryReply::Exception::KeysetDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const KeysetStateHistoryReply::Exception::InvalidHistoryInterval& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_keyset_state_flags(
            grpc::ServerContext*,
            const google::protobuf::Empty*,
            Api::Keyset::KeysetStateFlagsReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_keyset_state_flags requested");
            const auto info = get_keyset_state_flags_info();
            proto_wrap(info, _api_response);
            LOGGER.debug("server: answer = " + to_string(info));
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
};

class SearchKeysetService final : public Api::Keyset::SearchKeyset::SearchKeyset::Service
{
private:
    grpc::Status search_keyset(
            grpc::ServerContext*,
            const Api::Keyset::SearchKeysetRequest* _api_request,
            Api::Keyset::SearchKeysetReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: search_keyset requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("_api_request = " + request.to_string());
            const auto response = Fred::Registry::Keyset::search_keyset(request);
            proto_wrap(response, _api_response);
            LOGGER.debug("server: answer = " + response.to_string());
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status search_keyset_history(
            grpc::ServerContext*,
            const Api::Keyset::SearchKeysetHistoryRequest* _api_request,
            Api::Keyset::SearchKeysetHistoryReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: search_keyset_history requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: search_keyset_history(" + request.to_string() + ")");
            const auto response = Fred::Registry::Keyset::search_keyset(request);
            proto_wrap(response, _api_response);
            LOGGER.debug("server: answer = " + response.to_string());
            return grpc::Status::OK;
        }
        catch (const SearchKeysetHistoryReply::Exception::ChronologyViolation& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
};

}//namespace Fred::Registry::Keyset::{anonymous}

}//namespace Fred::Registry::Keyset

namespace Nsset {

namespace {

class NssetService final : public Api::Nsset::Nsset::Service
{
private:
    grpc::Status get_nsset_handle_history(
            grpc::ServerContext*,
            const Api::Nsset::NssetHandleHistoryRequest* _api_request,
            Api::Nsset::NssetHandleHistoryReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_nsset_handle_history requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: nsset_handle_history(" + request.to_string() + ")");
            const auto handle_history = nsset_handle_history(request);
            proto_wrap(handle_history, _api_response);
            LOGGER.debug("server: answer = " + handle_history.to_string());
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_nsset_history(
            grpc::ServerContext*,
            const Api::Nsset::NssetHistoryRequest* _api_request,
            Api::Nsset::NssetHistoryReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_nsset_history requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: nsset_history(" + request.to_string() + ")");
            const auto data_history = nsset_data_history(request);
            proto_wrap(data_history, _api_response);
            LOGGER.debug("server: answer = " + data_history.to_string());
            return grpc::Status::OK;
        }
        catch (const NssetDataHistoryReply::Exception::NssetDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const NssetDataHistoryReply::Exception::InvalidHistoryInterval& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_nsset_info(
            grpc::ServerContext*,
            const Api::Nsset::NssetInfoRequest* _api_request,
            Api::Nsset::NssetInfoReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_nsset_info requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: nsset_info(" + request.to_string() + ")");
            const auto info = nsset_info(request);
            proto_wrap(info, _api_response);
            LOGGER.debug("server: answer = " + info.to_string());
            return grpc::Status::OK;
        }
        catch (const NssetInfoReply::Exception::NssetDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_nsset_id(
            grpc::ServerContext*,
            const Api::Nsset::NssetIdRequest* _api_request,
            Api::Nsset::NssetIdReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_nsset_id requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: get_nsset_id(" + request.to_string() + ")");
            const auto id = Fred::Registry::Nsset::get_nsset_id(request);
            proto_wrap(id, _api_response);
            LOGGER.debug("server: answer = " + id.to_string());
            return grpc::Status::OK;
        }
        catch (const NssetIdReply::Exception::NssetDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_nsset_state(
            grpc::ServerContext*,
            const Api::Nsset::NssetStateRequest* _api_request,
            Api::Nsset::NssetStateReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_nsset_state requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: nsset_state(" + request.to_string() + ")");
            const auto state = nsset_state(request);
            proto_wrap(state, _api_response);
            LOGGER.debug("server: answer = " + state.to_string());
            return grpc::Status::OK;
        }
        catch (const NssetStateReply::Exception::NssetDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_nsset_state_history(
            grpc::ServerContext*,
            const Api::Nsset::NssetStateHistoryRequest* _api_request,
            Api::Nsset::NssetStateHistoryReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_nsset_state_history requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: nsset_state_history(" + request.to_string() + ")");
            const auto state_history = nsset_state_history(request);
            proto_wrap(state_history, _api_response);
            LOGGER.debug("server: answer = " + state_history.to_string());
            return grpc::Status::OK;
        }
        catch (const NssetStateHistoryReply::Exception::NssetDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const NssetStateHistoryReply::Exception::InvalidHistoryInterval& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_nsset_state_flags(
            grpc::ServerContext*,
            const google::protobuf::Empty*,
            Api::Nsset::NssetStateFlagsReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_nsset_state_flags requested");
            const auto info = get_nsset_state_flags_info();
            proto_wrap(info, _api_response);
            LOGGER.debug("server: answer = " + to_string(info));
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
};

class SearchNssetService final : public Api::Nsset::SearchNsset::SearchNsset::Service
{
private:
    grpc::Status search_nsset(
            grpc::ServerContext*,
            const Api::Nsset::SearchNssetRequest* _api_request,
            Api::Nsset::SearchNssetReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: search_nsset requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: search_nsset(" + request.to_string() + ")");
            const auto response = Fred::Registry::Nsset::search_nsset(request);
            proto_wrap(response, _api_response);
            LOGGER.debug("server: answer = " + response.to_string());
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status search_nsset_history(
            grpc::ServerContext*,
            const Api::Nsset::SearchNssetHistoryRequest* _api_request,
            Api::Nsset::SearchNssetHistoryReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: search_nsset_history requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: search_nsset_history(" + request.to_string() + ")");
            const auto response = Fred::Registry::Nsset::search_nsset(request);
            proto_wrap(response, _api_response);
            LOGGER.debug("server: answer = " + response.to_string());
            return grpc::Status::OK;
        }
        catch (const SearchNssetHistoryReply::Exception::ChronologyViolation& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
};

}//namespace Fred::Registry::Nsset::{anonymous}

}//namespace Fred::Registry::Nsset

namespace Registrar {

namespace {

class RegistrarService final : public Api::Registrar::Registrar::Service
{
private:
    grpc::Status get_registrar_credit(
            grpc::ServerContext*,
            const Api::Registrar::RegistrarCreditRequest* _api_request,
            Api::Registrar::RegistrarCreditReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_registrar_credit requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: registrar_credit(" + request.to_string() + ")");
            const auto info = registrar_credit(request);
            proto_wrap(info, _api_response);
            LOGGER.debug("server: answer = " + info.to_string());
            return grpc::Status::OK;
        }
        catch (const RegistrarCreditReply::Exception::RegistrarDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_registrar_info(
            grpc::ServerContext*,
            const Api::Registrar::RegistrarInfoRequest* _api_request,
            Api::Registrar::RegistrarInfoReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_registrar_info requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: registrar_info(" + request.to_string() + ")");
            const auto info = registrar_info(request);
            proto_wrap(info, _api_response);
            LOGGER.debug("server: answer = " + info.to_string());
            return grpc::Status::OK;
        }
        catch (const RegistrarInfoReply::Exception::RegistrarDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_registrar_handles(
            grpc::ServerContext*,
            const google::protobuf::Empty*,
            Api::Registrar::RegistrarListReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_registrar_handles requested");
            LOGGER.debug("server: registrar_list()");
            const auto info = registrar_list();
            proto_wrap(info, _api_response);
            LOGGER.debug("server: answer = " + info.to_string());
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
    grpc::Status get_registrar_zone_access(
            grpc::ServerContext*,
            const Api::Registrar::RegistrarZoneAccessRequest* _api_request,
            Api::Registrar::RegistrarZoneAccessReply* _api_response) override
    {
        LOGGING_CONTEXT(log_context);
        try
        {
            LOGGER.info("server: get_registrar_zone_access requested");
            const auto request = proto_unwrap(*_api_request);
            LOGGER.debug("server: registrar_zone_access(" + request.to_string() + ")");
            const auto info = registrar_zone_access(request);
            proto_wrap(info, _api_response);
            LOGGER.debug("server: answer = " + info.to_string());
            return grpc::Status::OK;
        }
        catch (const RegistrarZoneAccessReply::Exception::RegistrarDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _api_response);
            return grpc::Status::OK;
        }
        catch (const InternalServerError& e)
        {
            return on_internal_server_error(static_cast<const std::exception&>(e));
        }
        catch (const std::exception& e)
        {
            return on_std_exception(e);
        }
        catch (...)
        {
            return on_unknown_exception();
        }
    }
};

}//namespace Fred::Registry::Registrar::{anonymous}

}//namespace Fred::Registry::Registrar

namespace {

void run_server(const std::string& server_address)
{
    ::sigset_t blocked_signals;

    // Other threads created by grpc::Server will inherit a copy of the signal mask.
    ::sigemptyset(&blocked_signals);
    ::sigaddset(&blocked_signals, SIGHUP);
    ::sigaddset(&blocked_signals, SIGINT);
    ::sigaddset(&blocked_signals, SIGQUIT);
    ::sigaddset(&blocked_signals, SIGTERM);
    const auto pthread_sigmask_result = ::pthread_sigmask(SIG_BLOCK, &blocked_signals, nullptr);
    constexpr int operation_success = 0;
    if (pthread_sigmask_result != operation_success)
    {
        throw std::runtime_error(std::strerror(pthread_sigmask_result));
    }

    auto diagnostics_service =
            LibDiagnostics::make_grpc_service(
                    Fred::Registry::Diagnostics::about,
                    Fred::Registry::Diagnostics::status);

    Registry::Contact::ContactService contact_service;
    Registry::Domain::AdminService domain_admin_service;
    Registry::Domain::DomainService domain_service;
    Registry::Keyset::KeysetService keyset_service;
    Registry::Nsset::NssetService nsset_service;
    Registry::Registrar::RegistrarService registrar_service;

    Registry::Contact::SearchContactService search_contact_service;
    Registry::Domain::SearchDomainService search_domain_service;
    Registry::Keyset::SearchKeysetService search_keyset_service;
    Registry::Nsset::SearchNssetService search_nsset_service;

    grpc::ServerBuilder builder;

    // Listen on the given address without any authentication mechanism.
    builder.AddListeningPort(server_address, grpc::InsecureServerCredentials());

    // Register "service" as the instance through which we'll communicate with
    // clients. In this case it corresponds to an *synchronous* service.
    builder.RegisterService(diagnostics_service.get());
    builder.RegisterService(&contact_service);
    builder.RegisterService(&domain_admin_service);
    builder.RegisterService(&domain_service);
    builder.RegisterService(&keyset_service);
    builder.RegisterService(&nsset_service);
    builder.RegisterService(&registrar_service);

    builder.RegisterService(&search_contact_service);
    builder.RegisterService(&search_domain_service);
    builder.RegisterService(&search_keyset_service);
    builder.RegisterService(&search_nsset_service);

    // Finally assemble the server.
    std::unique_ptr<grpc::Server> server(builder.BuildAndStart());
    LOGGER.info("server: listen on " + server_address);

    int received_signal;
    const auto sigwait_result = ::sigwait(&blocked_signals, &received_signal);
    if (sigwait_result != operation_success)
    {
        throw std::runtime_error(std::strerror(sigwait_result));
    }
    switch (received_signal)
    {
        case SIGHUP:
            LOGGER.debug("SIGHUP received");
            break;
        case SIGINT:
            LOGGER.debug("SIGINT received");
            break;
        case SIGQUIT:
            LOGGER.debug("SIGQUIT received");
            break;
        case SIGTERM:
            LOGGER.debug("SIGTERM received");
            break;
        default:
            throw std::runtime_error("unexpected signal received");
    }
    const std::chrono::system_clock::time_point deadline = std::chrono::system_clock::now() + std::chrono::seconds(10);
    server->Shutdown(deadline);
}

}//namespace Fred::Registry::{anonymous}

}//namespace Fred::Registry
}//namespace Fred

namespace {

enum class LogHandler
{
    is_set,
    is_not_set
};

class SetLogHandler : public boost::static_visitor<LogHandler>
{
public:
    explicit SetLogHandler(Logging::Log& logger_instance) : logger_instance_(logger_instance) { }
    LogHandler operator()(const boost::blank&) const
    {
        return LogHandler::is_not_set;
    }
    LogHandler operator()(const Cfg::Options::Console& option) const
    {
        Logging::add_console_device(logger_instance_, option.min_severity);
        return LogHandler::is_set;
    }
    LogHandler operator()(const Cfg::Options::LogFile& option) const
    {
        Logging::add_file_device(logger_instance_, option.file_name, option.min_severity);
        return LogHandler::is_set;
    }
    LogHandler operator()(const Cfg::Options::SysLog& option) const
    {
        Logging::add_syslog_device(logger_instance_, option.facility, option.min_severity);
        return LogHandler::is_set;
    }
private:
    Logging::Log& logger_instance_;
};

std::string make_connect_string(const Cfg::Options::Database& option)
{
    std::string result;
    if (option.host != boost::none)
    {
        result = "host=" + *option.host;
    }
    if (option.port != boost::none)
    {
        if (!result.empty())
        {
            result += " ";
        }
        result += "port=" + std::to_string(*option.port);
    }
    if (option.user != boost::none)
    {
        if (!result.empty())
        {
            result += " ";
        }
        result += "user=" + *option.user;
    }
    if (option.dbname != boost::none)
    {
        if (!result.empty())
        {
            result += " ";
        }
        result += "dbname=" + *option.dbname;
    }
    if (option.password != boost::none)
    {
        if (!result.empty())
        {
            result += " ";
        }
        result += "password=" + *option.password;
    }
    return result;
}

}//namespace::{anonymous}

int main(int argc, char** argv)
{
    std::unique_ptr<Fred::Registry::LogContext> log_context;
    try
    {
        Cfg::Options::init(argc, argv);
    }
    catch (const Cfg::AllDone& e)
    {
        std::cout << e.what() << std::endl;
        return EXIT_SUCCESS;
    }
    catch (const Cfg::UnknownOption& unknown_option)
    {
        std::cerr << unknown_option.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (const Cfg::MissingOption& missing_option)
    {
        std::cerr << missing_option.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (const Cfg::Exception& e)
    {
        std::cerr << "Configuration problem: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Exception caught: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (...)
    {
        std::cerr << "Unexpected exception caught" << std::endl;
        return EXIT_FAILURE;
    }

    switch (boost::apply_visitor(SetLogHandler(LOGGER), Cfg::Options::get().log))
    {
        case LogHandler::is_set:
            log_context = std::make_unique<Fred::Registry::LogContext>("main");
            LOGGER.debug("fred-registry-services configuration successfully loaded");
            break;
        case LogHandler::is_not_set:
            std::cout << "fred-registry-services configuration successfully loaded" << std::endl;
            break;
    }
    Database::emplace_default_manager<Database::StandaloneManager>(
            make_connect_string(Cfg::Options::get().database));

    try
    {
        Fred::Registry::run_server(Cfg::Options::get().registry.listen_on);
        LOGGER.info("Server shutdown");
        return EXIT_SUCCESS;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Exception caught: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (...)
    {
        std::cerr << "Unexpected exception caught" << std::endl;
        return EXIT_FAILURE;
    }
}
