/*
 * Copyright (C) 2019-2022  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/proto_unwrap.hh"

#include "src/cfg.hh"
#include "src/util/deletable.hh"
#include "src/util/into.hh"

#include "util/log/logger.hh"

#include <boost/uuid/string_generator.hpp>
#include <boost/uuid/uuid_io.hpp>

#include <algorithm>
#include <exception>

namespace Fred {
namespace Registry {

template <>
struct ProtoTypeTraits<std::string>
{
    using RegistryType = std::string;
};

template <typename T>
typename ProtoTypeTraits<T>::RegistryType proto_unwrap(const T& src)
{
    static_assert(std::is_same<typename ProtoTypeTraits<T>::RegistryType, T>::value, "no conversion available");
    return src;
}

template <>
struct ProtoTypeTraits<Registry::Api::Uuid>
{
    using RegistryType = boost::uuids::uuid;
};

template <>
struct ProtoTypeTraits<Api::Contact::ContactId>
{
    using RegistryType = Contact::ContactId;
};

template <>
struct ProtoTypeTraits<Api::Contact::ContactHistoryId>
{
    using RegistryType = Contact::ContactHistoryId;
};

template <>
struct ProtoTypeTraits<Api::Contact::ContactHandle>
{
    using RegistryType = Contact::ContactHandle;
};

template <>
struct ProtoTypeTraits<Api::Contact::ContactHistoryInterval>
{
    using RegistryType = Contact::ContactHistoryInterval;
};

template <>
struct ProtoTypeTraits<Api::Domain::DomainId>
{
    using RegistryType = Domain::DomainId;
};

template <>
struct ProtoTypeTraits<Api::Domain::DomainHistoryId>
{
    using RegistryType = Domain::DomainHistoryId;
};

template <>
struct ProtoTypeTraits<Api::Domain::Fqdn>
{
    using RegistryType = Domain::Fqdn;
};

template <>
struct ProtoTypeTraits<Api::Domain::DomainHistoryInterval>
{
    using RegistryType = Domain::DomainHistoryInterval;
};

template <>
struct ProtoTypeTraits<Api::Domain::DomainStateFlagRequest>
{
    using RegistryType = Domain::DomainStateFlagRequest;
};

template <>
struct ProtoTypeTraits<Api::Domain::UpdateDomainsDeleteAdditionalNotifyInfoRequest::DomainAdditionalInfo>
{
    using RegistryType = Domain::UpdateDomainsAdditionalNotifyInfoRequest::DomainAdditionalInfo;
};

template <>
struct ProtoTypeTraits<Api::Domain::UpdateDomainsOutzoneAdditionalNotifyInfoRequest::DomainAdditionalInfo>
{
    using RegistryType = Domain::UpdateDomainsAdditionalNotifyInfoRequest::DomainAdditionalInfo;
};

template <>
struct ProtoTypeTraits<Api::Keyset::KeysetId>
{
    using RegistryType = Keyset::KeysetId;
};

template <>
struct ProtoTypeTraits<Api::Keyset::KeysetHistoryId>
{
    using RegistryType = Keyset::KeysetHistoryId;
};

template <>
struct ProtoTypeTraits<Api::Keyset::KeysetHandle>
{
    using RegistryType = Keyset::KeysetHandle;
};

template <>
struct ProtoTypeTraits<Api::Keyset::KeysetHistoryInterval>
{
    using RegistryType = Keyset::KeysetHistoryInterval;
};

template <>
struct ProtoTypeTraits<Api::Nsset::NssetId>
{
    using RegistryType = Nsset::NssetId;
};

template <>
struct ProtoTypeTraits<Api::Nsset::NssetHistoryId>
{
    using RegistryType = Nsset::NssetHistoryId;
};

template <>
struct ProtoTypeTraits<Api::Nsset::NssetHandle>
{
    using RegistryType = Nsset::NssetHandle;
};

template <>
struct ProtoTypeTraits<Api::Nsset::NssetHistoryInterval>
{
    using RegistryType = Nsset::NssetHistoryInterval;
};

template <>
boost::uuids::uuid proto_unwrap(const Registry::Api::Uuid& src)
{
    return boost::uuids::string_generator()(src.value());
}

template <>
Contact::ContactId proto_unwrap(const Api::Contact::ContactId& src)
{
    return Util::make_strong<Contact::ContactId>(proto_unwrap(src.uuid()));
}

template <>
Contact::ContactHistoryId proto_unwrap(const Api::Contact::ContactHistoryId& src)
{
    return Util::make_strong<Contact::ContactHistoryId>(proto_unwrap(src.uuid()));
}

template <>
Contact::ContactInfoRequest proto_unwrap(const Api::Contact::ContactInfoRequest& src)
{
    Contact::ContactInfoRequest dst;
    dst.contact_id = proto_unwrap(src.contact_id());
    if (src.has_contact_history_id())
    {
        dst.contact_history_id = proto_unwrap(src.contact_history_id());
    }
    return dst;
}

template <>
Contact::ContactHandle proto_unwrap(const Api::Contact::ContactHandle& src)
{
    return Util::make_strong<Contact::ContactHandle>(proto_unwrap(src.value()));
}

template <>
struct ProtoTypeTraits<Api::PaginationRequest>
{
    using RegistryType = PaginationRequest;
};

template <>
Contact::ContactIdRequest proto_unwrap(const Api::Contact::ContactIdRequest& src)
{
    Contact::ContactIdRequest dst;
    dst.contact_handle = proto_unwrap(src.contact_handle());
    return dst;
}

namespace {

Contact::ContactHistoryInterval::HistoryId unwrap_contact_history_interval_history_id(const Api::Contact::ContactHistoryId& src)
{
    return Util::make_strong<Contact::ContactHistoryInterval::HistoryId>(
            Util::get_raw_value_from(proto_unwrap(src)));
}

std::chrono::time_point<std::chrono::system_clock, std::chrono::nanoseconds> unwrap_system_clock_nanoseconds(
        const google::protobuf::Timestamp& src)
{
    const std::chrono::time_point<std::chrono::system_clock, std::chrono::nanoseconds> dst(
            std::chrono::nanoseconds(1'000'000'000 * src.seconds() + src.nanos()));
    return dst;
}

}//namespace Fred::Registry::{anonymous}

template <>
Contact::ContactHistoryInterval proto_unwrap(const Api::Contact::ContactHistoryInterval& src)
{
    auto lower_limit = Contact::ContactHistoryInterval::LowerLimit(Contact::ContactHistoryInterval::NoLimit());
    if (src.has_lower_limit())
    {
        if (src.lower_limit().has_contact_history_id())
        {
            lower_limit.value = unwrap_contact_history_interval_history_id(src.lower_limit().contact_history_id());
        }
        else if (src.lower_limit().has_timestamp())
        {
            lower_limit.value = unwrap_system_clock_nanoseconds(src.lower_limit().timestamp());
        }
    }
    auto upper_limit = Contact::ContactHistoryInterval::UpperLimit(Contact::ContactHistoryInterval::NoLimit());
    if (src.has_upper_limit())
    {
        if (src.upper_limit().has_contact_history_id())
        {
            upper_limit.value = unwrap_contact_history_interval_history_id(src.upper_limit().contact_history_id());
        }
        else if (src.upper_limit().has_timestamp())
        {
            upper_limit.value = unwrap_system_clock_nanoseconds(src.upper_limit().timestamp());
        }
    }
    return Contact::ContactHistoryInterval(lower_limit, upper_limit);
}

namespace {

decltype(auto) get_disclose_flag_operation(bool to_publish)
{
    return to_publish ? Contact::publicly_available_data()
                      : !Contact::publicly_available_data();
}

decltype(auto) unwrap_set_email_addresses(decltype(std::declval<const Api::Contact::UpdateContactRequest&>().set_emails()) src)
{
    Contact::SetEmailAddresses dst;
    dst.values.reserve(src.values().size());
    for (const auto& email : src.values())
    {
        dst.values.emplace_back(Util::make_strong<Contact::EmailAddress>(email.value()));
    }
    return dst;
}

decltype(auto) unwrap_set_additional_identifier(const Api::Contact::ContactAdditionalIdentifier& src)
{
    if (src.has_national_identity_number())
    {
        return Util::to_set(Contact::ContactAdditionalIdentifier{
                Util::make_strong<Contact::NationalIdentityNumber>(
                        src.national_identity_number().value())});
    }
    if (src.has_national_identity_card())
    {
        return Util::to_set(Contact::ContactAdditionalIdentifier{
                Util::make_strong<Contact::NationalIdentityCard>(
                        src.national_identity_card().value())});
    }
    if (src.has_passport_number())
    {
        return Util::to_set(Contact::ContactAdditionalIdentifier{
                Util::make_strong<Contact::PassportNumber>(
                        src.passport_number().value())});
    }
    if (src.has_company_registration_number())
    {
        return Util::to_set(Contact::ContactAdditionalIdentifier{
                Util::make_strong<Contact::CompanyRegistrationNumber>(
                        src.company_registration_number().value())});
    }
    if (src.has_social_security_number())
    {
        return Util::to_set(Contact::ContactAdditionalIdentifier{
                Util::make_strong<Contact::SocialSecurityNumber>(
                        src.social_security_number().value())});
    }
    if (src.has_birthdate())
    {
        return Util::to_set(Contact::ContactAdditionalIdentifier{
                Util::make_strong<Contact::Birthdate>(
                        src.birthdate().value())});
    }
    return Util::to_delete<Contact::ContactAdditionalIdentifier>();
}

}//namespace Fred::Registry::{anonymous}

template <>
Contact::PlaceAddress proto_unwrap(const Api::PlaceAddress& src)
{
    Contact::PlaceAddress dst;
    dst.street.reserve(src.street().size());
    for (const auto& street : src.street())
    {
        dst.street.emplace_back(street);
    }
    dst.city = src.city();
    dst.state_or_province = src.state_or_province();
    dst.postal_code = Util::make_strong<Contact::PostalCode>(src.postal_code().value());
    dst.country_code = Util::make_strong<Contact::CountryCode>(src.country_code().value());
    return dst;
}

template <>
Contact::ContactAddress proto_unwrap(const Api::Contact::ContactAddress& src)
{
    Contact::ContactAddress dst;
    dst.company = src.company();
    dst.street.reserve(src.street().size());
    for (const auto& street : src.street())
    {
        dst.street.emplace_back(street);
    }
    dst.city = src.city();
    dst.state_or_province = src.state_or_province();
    dst.postal_code = Util::make_strong<Contact::PostalCode>(src.postal_code().value());
    dst.country_code = Util::make_strong<Contact::CountryCode>(src.country_code().value());
    return dst;
}

template <>
Contact::WarningLetter proto_unwrap(const Api::Contact::WarningLetter& src)
{
    Contact::WarningLetter dst;
    switch (src.preference())
    {
        case Api::Contact::WarningLetter_SendingPreference_Enum_to_send:
            dst.preference = Contact::WarningLetter::SendingPreference::to_send;
            return dst;
        case Api::Contact::WarningLetter_SendingPreference_Enum_not_to_send:
            dst.preference = Contact::WarningLetter::SendingPreference::not_to_send;
            return dst;
        case Api::Contact::WarningLetter_SendingPreference_Enum_not_specified:
            dst.preference = Contact::WarningLetter::SendingPreference::not_specified;
            return dst;
        default:
        {
            struct UnknownPreference : std::exception
            {
                const char* what() const noexcept override { return "unknown preference"; }
            };
            throw UnknownPreference{};
        }
    }
}

template <>
Contact::AuthInfo proto_unwrap(const Api::AuthInfo& src)
{
    Contact::AuthInfo dst;
    dst.password = src.password();
    return dst;
}

template <>
Contact::UpdateContactRequest proto_unwrap(const Api::Contact::UpdateContactRequest& src)
{
    Contact::UpdateContactRequest dst;
    dst.contact_id = proto_unwrap(src.contact_id());
    dst.contact_history_id = proto_unwrap(src.contact_history_id());
    if (src.SetName_case() == Api::Contact::UpdateContactRequest::kSetName)
    {
        dst.set_name = src.set_name();
    }
    if (src.SetOrganization_case() == Api::Contact::UpdateContactRequest::kSetOrganization)
    {
        dst.set_organization = src.set_organization();
    }
    if (src.has_set_place())
    {
        dst.set_place = proto_unwrap(src.set_place());
    }
    if (src.has_set_telephone())
    {
        dst.set_telephone = Util::make_strong<Contact::PhoneNumber>(src.set_telephone().value());
    }
    if (src.has_set_fax())
    {
        dst.set_fax = Util::make_strong<Contact::PhoneNumber>(src.set_fax().value());
    }
    if (src.has_set_emails())
    {
        dst.set_emails = unwrap_set_email_addresses(src.set_emails());
    }
    if (src.has_set_notify_emails())
    {
        dst.set_notify_emails = unwrap_set_email_addresses(src.set_notify_emails());
    }
    if (src.has_set_vat_identification_number())
    {
        dst.set_vat_identification_number = Util::make_strong<Contact::VatIdentificationNumber>(src.set_vat_identification_number().value());
    }
    if (src.SetContactAdditionalIdentifier_case() == Api::Contact::UpdateContactRequest::kSetAdditionalIdentifier)
    {
        dst.set_additional_identifier = unwrap_set_additional_identifier(src.set_additional_identifier());
    }
    if (src.SetMailingAddress_case() == Api::Contact::UpdateContactRequest::kSetMailingAddress)
    {
        dst.set_mailing_address = Util::to_set(proto_unwrap(src.set_mailing_address()));
    }
    else if (src.SetMailingAddress_case() == Api::Contact::UpdateContactRequest::kDeleteMailingAddress)
    {
        dst.set_mailing_address = Util::to_delete<Contact::ContactAddress>();
    }
    if (src.SetBillingAddress_case() == Api::Contact::UpdateContactRequest::kSetBillingAddress)
    {
        dst.set_billing_address = Util::to_set(proto_unwrap(src.set_billing_address()));
    }
    else if (src.SetBillingAddress_case() == Api::Contact::UpdateContactRequest::kDeleteBillingAddress)
    {
        dst.set_billing_address = Util::to_delete<Contact::ContactAddress>();
    }
    if (src.SetShippingAddress1_case() == Api::Contact::UpdateContactRequest::kSetShippingAddress1)
    {
        dst.set_shipping_address[0] = Util::to_set(proto_unwrap(src.set_shipping_address1()));
    }
    else if (src.SetShippingAddress1_case() == Api::Contact::UpdateContactRequest::kDeleteShippingAddress1)
    {
        dst.set_shipping_address[0] = Util::to_delete<Contact::ContactAddress>();
    }
    if (src.SetShippingAddress2_case() == Api::Contact::UpdateContactRequest::kSetShippingAddress2)
    {
        dst.set_shipping_address[1] = Util::to_set(proto_unwrap(src.set_shipping_address2()));
    }
    else if (src.SetShippingAddress2_case() == Api::Contact::UpdateContactRequest::kDeleteShippingAddress2)
    {
        dst.set_shipping_address[1] = Util::to_delete<Contact::ContactAddress>();
    }
    if (src.SetShippingAddress3_case() == Api::Contact::UpdateContactRequest::kSetShippingAddress3)
    {
        dst.set_shipping_address[2] = Util::to_set(proto_unwrap(src.set_shipping_address3()));
    }
    else if (src.SetShippingAddress3_case() == Api::Contact::UpdateContactRequest::kDeleteShippingAddress3)
    {
        dst.set_shipping_address[2] = Util::to_delete<Contact::ContactAddress>();
    }
    if (src.has_set_warning_letter())
    {
        dst.set_warning_letter = proto_unwrap(src.set_warning_letter());
    }
    if (src.has_set_auth_info())
    {
        if (!src.set_auth_info().password().empty())
        {
            dst.set_auth_info = Util::to_set(proto_unwrap(src.set_auth_info()));
        }
        else
        {
            dst.set_auth_info = Util::to_delete<Contact::AuthInfo>();
        }
    }
    for (const auto& set_publish : src.set_publish())
    {
        if (set_publish.first == "name")
        {
            dst.set_name_privacy = get_disclose_flag_operation(set_publish.second);
        }
        else if (set_publish.first == "organization")
        {
            dst.set_organization_privacy = get_disclose_flag_operation(set_publish.second);
        }
        else if (set_publish.first == "place")
        {
            dst.set_place_privacy = get_disclose_flag_operation(set_publish.second);
        }
        else if (set_publish.first == "telephone")
        {
            dst.set_telephone_privacy = get_disclose_flag_operation(set_publish.second);
        }
        else if (set_publish.first == "fax")
        {
            dst.set_fax_privacy = get_disclose_flag_operation(set_publish.second);
        }
        else if (set_publish.first == "emails")
        {
            dst.set_emails_privacy = get_disclose_flag_operation(set_publish.second);
        }
        else if (set_publish.first == "notify_emails")
        {
            dst.set_notify_emails_privacy = get_disclose_flag_operation(set_publish.second);
        }
        else if (set_publish.first == "vat_identification_number")
        {
            dst.set_vat_identification_number_privacy = get_disclose_flag_operation(set_publish.second);
        }
        else if (set_publish.first == "additional_identifier")
        {
            dst.set_additional_identifier_privacy = get_disclose_flag_operation(set_publish.second);
        }
        else
        {
            struct UnknownDiscloseFlagName : std::exception
            {
                const char* what() const noexcept override { return "unknown disclose flag name"; }
            };
            throw UnknownDiscloseFlagName{};
        }
    }
    if (src.has_log_entry_id())
    {
        dst.log_entry_id = Util::make_strong<LogEntryId>(src.log_entry_id().value());
    }
    const bool registrar_configured = Cfg::Options::get().fred.registrar_originator != boost::none;
    if (!registrar_configured)
    {
        struct NoRegistrarConfigured : std::exception
        {
            const char* what() const noexcept override { return "No registrar-originator configured"; }
        };
        throw NoRegistrarConfigured{};
    }
    dst.registrar_originator = Util::make_strong<Registrar::RegistrarHandle>(*Cfg::Options::get().fred.registrar_originator);
    return dst;
}

template <>
Contact::SearchContactRequest proto_unwrap(const Api::Contact::SearchContactRequest& src)
{
    Contact::SearchContactRequest dst;
    dst.query_values.reserve(src.query_values_size());
    for (const auto& value : src.query_values())
    {
        dst.query_values.push_back(value);
    }
    if (src.Limit_case() != Api::Contact::SearchContactRequest::LIMIT_NOT_SET)
    {
        dst.limit = src.limit();
    }
    else
    {
        dst.limit = boost::none;
    }
    for (const auto& item : src.searched_items())
    {
        if (item == Util::Into<std::string>::from(Contact::ContactItem::contact_handle))
        {
            dst.searched_items.insert(Contact::ContactItem::contact_handle);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::name))
        {
            dst.searched_items.insert(Contact::ContactItem::name);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::organization))
        {
            dst.searched_items.insert(Contact::ContactItem::organization);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::place_street))
        {
            dst.searched_items.insert(Contact::ContactItem::place_street);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::place_city))
        {
            dst.searched_items.insert(Contact::ContactItem::place_city);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::place_state_or_province))
        {
            dst.searched_items.insert(Contact::ContactItem::place_state_or_province);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::place_postal_code))
        {
            dst.searched_items.insert(Contact::ContactItem::place_postal_code);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::telephone))
        {
            dst.searched_items.insert(Contact::ContactItem::telephone);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::fax))
        {
            dst.searched_items.insert(Contact::ContactItem::fax);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::email))
        {
            dst.searched_items.insert(Contact::ContactItem::email);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::notify_email))
        {
            dst.searched_items.insert(Contact::ContactItem::notify_email);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::vat_identification_number))
        {
            dst.searched_items.insert(Contact::ContactItem::vat_identification_number);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::additional_identifier))
        {
            dst.searched_items.insert(Contact::ContactItem::additional_identifier);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_company))
        {
            dst.searched_items.insert(Contact::ContactItem::address_company);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_street))
        {
            dst.searched_items.insert(Contact::ContactItem::address_street);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_city))
        {
            dst.searched_items.insert(Contact::ContactItem::address_city);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_state_or_province))
        {
            dst.searched_items.insert(Contact::ContactItem::address_state_or_province);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_postal_code))
        {
            dst.searched_items.insert(Contact::ContactItem::address_postal_code);
        }
        else
        {
            LOGGER.warning("unknown item \"" + item + "\" in SearchContactRequest");
        }
    }
    if (dst.searched_items.empty())
    {
        dst.searched_items.insert(Contact::ContactItem::contact_handle);
        dst.searched_items.insert(Contact::ContactItem::name);
        dst.searched_items.insert(Contact::ContactItem::organization);
        dst.searched_items.insert(Contact::ContactItem::place_street);
        dst.searched_items.insert(Contact::ContactItem::place_city);
        dst.searched_items.insert(Contact::ContactItem::place_state_or_province);
        dst.searched_items.insert(Contact::ContactItem::place_postal_code);
        dst.searched_items.insert(Contact::ContactItem::telephone);
        dst.searched_items.insert(Contact::ContactItem::fax);
        dst.searched_items.insert(Contact::ContactItem::email);
        dst.searched_items.insert(Contact::ContactItem::notify_email);
        dst.searched_items.insert(Contact::ContactItem::vat_identification_number);
        dst.searched_items.insert(Contact::ContactItem::additional_identifier);
        dst.searched_items.insert(Contact::ContactItem::address_company);
        dst.searched_items.insert(Contact::ContactItem::address_street);
        dst.searched_items.insert(Contact::ContactItem::address_city);
        dst.searched_items.insert(Contact::ContactItem::address_state_or_province);
        dst.searched_items.insert(Contact::ContactItem::address_postal_code);
    }
    return dst;
}

template <>
Contact::SearchContactHistoryRequest proto_unwrap(const Api::Contact::SearchContactHistoryRequest& src)
{
    Contact::SearchContactHistoryRequest dst;
    dst.query_values.reserve(src.query_values_size());
    for (const auto& value : src.query_values())
    {
        dst.query_values.push_back(value);
    }
    if (src.Limit_case() != Api::Contact::SearchContactHistoryRequest::LIMIT_NOT_SET)
    {
        dst.max_number_of_contacts = src.max_number_of_contacts();
    }
    else
    {
        dst.max_number_of_contacts = boost::none;
    }
    if (src.has_data_valid_from())
    {
        dst.data_valid_from = unwrap_system_clock_nanoseconds(src.data_valid_from());
    }
    else
    {
        dst.data_valid_from = boost::none;
    }
    if (src.has_data_valid_to())
    {
        dst.data_valid_to = unwrap_system_clock_nanoseconds(src.data_valid_to());
    }
    else
    {
        dst.data_valid_to = boost::none;
    }
    for (const auto& item : src.searched_items())
    {
        if (item == Util::Into<std::string>::from(Contact::ContactItem::contact_handle))
        {
            dst.searched_items.insert(Contact::ContactItem::contact_handle);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::name))
        {
            dst.searched_items.insert(Contact::ContactItem::name);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::organization))
        {
            dst.searched_items.insert(Contact::ContactItem::organization);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::place_street))
        {
            dst.searched_items.insert(Contact::ContactItem::place_street);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::place_city))
        {
            dst.searched_items.insert(Contact::ContactItem::place_city);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::place_state_or_province))
        {
            dst.searched_items.insert(Contact::ContactItem::place_state_or_province);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::place_postal_code))
        {
            dst.searched_items.insert(Contact::ContactItem::place_postal_code);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::telephone))
        {
            dst.searched_items.insert(Contact::ContactItem::telephone);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::fax))
        {
            dst.searched_items.insert(Contact::ContactItem::fax);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::email))
        {
            dst.searched_items.insert(Contact::ContactItem::email);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::notify_email))
        {
            dst.searched_items.insert(Contact::ContactItem::notify_email);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::vat_identification_number))
        {
            dst.searched_items.insert(Contact::ContactItem::vat_identification_number);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::additional_identifier))
        {
            dst.searched_items.insert(Contact::ContactItem::additional_identifier);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_company))
        {
            dst.searched_items.insert(Contact::ContactItem::address_company);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_street))
        {
            dst.searched_items.insert(Contact::ContactItem::address_street);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_city))
        {
            dst.searched_items.insert(Contact::ContactItem::address_city);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_state_or_province))
        {
            dst.searched_items.insert(Contact::ContactItem::address_state_or_province);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_postal_code))
        {
            dst.searched_items.insert(Contact::ContactItem::address_postal_code);
        }
        else
        {
            LOGGER.warning("unknown item \"" + item + "\" in SearchContactHistoryRequest");
        }
    }
    if (dst.searched_items.empty())
    {
        dst.searched_items.insert(Contact::ContactItem::contact_handle);
        dst.searched_items.insert(Contact::ContactItem::name);
        dst.searched_items.insert(Contact::ContactItem::organization);
        dst.searched_items.insert(Contact::ContactItem::place_street);
        dst.searched_items.insert(Contact::ContactItem::place_city);
        dst.searched_items.insert(Contact::ContactItem::place_state_or_province);
        dst.searched_items.insert(Contact::ContactItem::place_postal_code);
        dst.searched_items.insert(Contact::ContactItem::telephone);
        dst.searched_items.insert(Contact::ContactItem::fax);
        dst.searched_items.insert(Contact::ContactItem::email);
        dst.searched_items.insert(Contact::ContactItem::notify_email);
        dst.searched_items.insert(Contact::ContactItem::vat_identification_number);
        dst.searched_items.insert(Contact::ContactItem::additional_identifier);
        dst.searched_items.insert(Contact::ContactItem::address_company);
        dst.searched_items.insert(Contact::ContactItem::address_street);
        dst.searched_items.insert(Contact::ContactItem::address_city);
        dst.searched_items.insert(Contact::ContactItem::address_state_or_province);
        dst.searched_items.insert(Contact::ContactItem::address_postal_code);
    }
    return dst;
}

template <>
Contact::ContactDataHistoryRequest proto_unwrap(const Api::Contact::ContactHistoryRequest& src)
{
    return Contact::ContactDataHistoryRequest(proto_unwrap(src.contact_id()),
                                              proto_unwrap(src.history()));
}

template <>
Contact::ContactHandleHistoryRequest proto_unwrap(const Api::Contact::ContactHandleHistoryRequest& src)
{
    return Contact::ContactHandleHistoryRequest(proto_unwrap(src.contact_handle()));
}

template <>
Contact::ContactStateRequest proto_unwrap(const Api::Contact::ContactStateRequest& src)
{
    Contact::ContactStateRequest dst;
    dst.contact_id = proto_unwrap(src.contact_id());
    return dst;
}

template <>
Contact::ContactStateHistoryRequest proto_unwrap(const Api::Contact::ContactStateHistoryRequest& src)
{
    return Contact::ContactStateHistoryRequest(proto_unwrap(src.contact_id()), proto_unwrap(src.history()));
}

template <>
Keyset::KeysetId proto_unwrap(const Api::Keyset::KeysetId& src)
{
    return Util::make_strong<Keyset::KeysetId>(proto_unwrap(src.uuid()));
}

template <>
Keyset::KeysetHistoryId proto_unwrap(const Api::Keyset::KeysetHistoryId& src)
{
    return Util::make_strong<Keyset::KeysetHistoryId>(proto_unwrap(src.uuid()));
}

template <>
Keyset::KeysetInfoRequest proto_unwrap(const Api::Keyset::KeysetInfoRequest& src)
{
    Keyset::KeysetInfoRequest dst;
    dst.keyset_id = proto_unwrap(src.keyset_id());
    if (src.has_keyset_history_id())
    {
        dst.keyset_history_id = proto_unwrap(src.keyset_history_id());
    }
    return dst;
}

template <>
Keyset::KeysetHandle proto_unwrap(const Api::Keyset::KeysetHandle& src)
{
    return Util::make_strong<Keyset::KeysetHandle>(proto_unwrap(src.value()));
}

template <>
Keyset::KeysetIdRequest proto_unwrap(const Api::Keyset::KeysetIdRequest& src)
{
    Keyset::KeysetIdRequest dst;
    dst.keyset_handle = proto_unwrap(src.keyset_handle());
    return dst;
}

namespace {

Keyset::KeysetHistoryInterval::HistoryId unwrap_keyset_history_interval_history_id(const Api::Keyset::KeysetHistoryId& src)
{
    return Util::make_strong<Keyset::KeysetHistoryInterval::HistoryId>(
            Util::get_raw_value_from(proto_unwrap(src)));
}

}//namespace Fred::Registry::{anonymous}

template <>
Keyset::KeysetHistoryInterval proto_unwrap(const Api::Keyset::KeysetHistoryInterval& src)
{
    auto lower_limit = Keyset::KeysetHistoryInterval::LowerLimit(Keyset::KeysetHistoryInterval::NoLimit());
    if (src.has_lower_limit())
    {
        if (src.lower_limit().has_keyset_history_id())
        {
            lower_limit.value = unwrap_keyset_history_interval_history_id(src.lower_limit().keyset_history_id());
        }
        else if (src.lower_limit().has_timestamp())
        {
            lower_limit.value = unwrap_system_clock_nanoseconds(src.lower_limit().timestamp());
        }
    }
    auto upper_limit = Keyset::KeysetHistoryInterval::UpperLimit(Keyset::KeysetHistoryInterval::NoLimit());
    if (src.has_upper_limit())
    {
        if (src.upper_limit().has_keyset_history_id())
        {
            upper_limit.value = unwrap_keyset_history_interval_history_id(src.upper_limit().keyset_history_id());
        }
        else if (src.upper_limit().has_timestamp())
        {
            upper_limit.value = unwrap_system_clock_nanoseconds(src.upper_limit().timestamp());
        }
    }
    return Keyset::KeysetHistoryInterval(lower_limit, upper_limit);
}

template <>
Keyset::SearchKeysetRequest proto_unwrap(const Api::Keyset::SearchKeysetRequest& src)
{
    Keyset::SearchKeysetRequest dst;
    dst.query_values.reserve(src.query_values_size());
    for (const auto& value : src.query_values())
    {
        dst.query_values.push_back(value);
    }
    if (src.Limit_case() != Api::Keyset::SearchKeysetRequest::LIMIT_NOT_SET)
    {
        dst.limit = src.limit();
    }
    else
    {
        dst.limit = boost::none;
    }
    for (const auto& item : src.searched_items())
    {
        if (item == Util::Into<std::string>::from(Keyset::KeysetItem::keyset_handle))
        {
            dst.searched_items.insert(Keyset::KeysetItem::keyset_handle);
        }
        else if (item == Util::Into<std::string>::from(Keyset::KeysetItem::public_key))
        {
            dst.searched_items.insert(Keyset::KeysetItem::public_key);
        }
        else
        {
            LOGGER.warning("unknown item \"" + item + "\" in SearchKeysetRequest");
        }
    }
    if (dst.searched_items.empty())
    {
        dst.searched_items = { Keyset::KeysetItem::keyset_handle,
                               Keyset::KeysetItem::public_key };
    }
    return dst;
}

template <>
Keyset::SearchKeysetHistoryRequest proto_unwrap(const Api::Keyset::SearchKeysetHistoryRequest& src)
{
    Keyset::SearchKeysetHistoryRequest dst;
    dst.query_values.reserve(src.query_values_size());
    for (const auto& value : src.query_values())
    {
        dst.query_values.push_back(value);
    }
    if (src.Limit_case() != Api::Keyset::SearchKeysetHistoryRequest::LIMIT_NOT_SET)
    {
        dst.max_number_of_keysets = src.max_number_of_keysets();
    }
    else
    {
        dst.max_number_of_keysets = boost::none;
    }
    if (src.has_data_valid_from())
    {
        dst.data_valid_from = unwrap_system_clock_nanoseconds(src.data_valid_from());
    }
    else
    {
        dst.data_valid_from = boost::none;
    }
    if (src.has_data_valid_to())
    {
        dst.data_valid_to = unwrap_system_clock_nanoseconds(src.data_valid_to());
    }
    else
    {
        dst.data_valid_to = boost::none;
    }
    for (const auto& item : src.searched_items())
    {
        if (item == Util::Into<std::string>::from(Keyset::KeysetItem::keyset_handle))
        {
            dst.searched_items.insert(Keyset::KeysetItem::keyset_handle);
        }
        else if (item == Util::Into<std::string>::from(Keyset::KeysetItem::public_key))
        {
            dst.searched_items.insert(Keyset::KeysetItem::public_key);
        }
        else
        {
            LOGGER.warning("unknown item \"" + item + "\" in SearchKeysetHistoryRequest");
        }
    }
    if (dst.searched_items.empty())
    {
        dst.searched_items = { Keyset::KeysetItem::keyset_handle,
                               Keyset::KeysetItem::public_key };
    }
    return dst;
}

template <>
Keyset::KeysetDataHistoryRequest proto_unwrap(const Api::Keyset::KeysetHistoryRequest& src)
{
    return Keyset::KeysetDataHistoryRequest(
            proto_unwrap(src.keyset_id()),
            proto_unwrap(src.history()));
}

template <>
Keyset::KeysetHandleHistoryRequest proto_unwrap(const Api::Keyset::KeysetHandleHistoryRequest& src)
{
    return Keyset::KeysetHandleHistoryRequest(proto_unwrap(src.keyset_handle()));
}

template <>
Keyset::KeysetStateRequest proto_unwrap(const Api::Keyset::KeysetStateRequest& src)
{
    Keyset::KeysetStateRequest dst;
    dst.keyset_id = proto_unwrap(src.keyset_id());
    return dst;
}

template <>
Keyset::KeysetStateHistoryRequest proto_unwrap(const Api::Keyset::KeysetStateHistoryRequest& src)
{
    return Keyset::KeysetStateHistoryRequest(proto_unwrap(src.keyset_id()), proto_unwrap(src.history()));
}

template <>
Nsset::NssetId proto_unwrap(const Api::Nsset::NssetId& src)
{
    return Util::make_strong<Nsset::NssetId>(proto_unwrap(src.uuid()));
}

template <>
Nsset::NssetHistoryId proto_unwrap(const Api::Nsset::NssetHistoryId& src)
{
    return Util::make_strong<Nsset::NssetHistoryId>(proto_unwrap(src.uuid()));
}

template <>
Nsset::NssetInfoRequest proto_unwrap(const Api::Nsset::NssetInfoRequest& src)
{
    Nsset::NssetInfoRequest dst;
    dst.nsset_id = proto_unwrap(src.nsset_id());
    if (src.has_nsset_history_id())
    {
        dst.nsset_history_id = proto_unwrap(src.nsset_history_id());
    }
    return dst;
}

template <>
Nsset::NssetHandle proto_unwrap(const Api::Nsset::NssetHandle& src)
{
    return Util::make_strong<Nsset::NssetHandle>(proto_unwrap(src.value()));
}

template <>
Nsset::NssetIdRequest proto_unwrap(const Api::Nsset::NssetIdRequest& src)
{
    Nsset::NssetIdRequest dst;
    dst.nsset_handle = proto_unwrap(src.nsset_handle());
    return dst;
}

namespace {

Nsset::NssetHistoryInterval::HistoryId unwrap_nsset_history_interval_history_id(const Api::Nsset::NssetHistoryId& src)
{
    return Util::make_strong<Nsset::NssetHistoryInterval::HistoryId>(
            Util::get_raw_value_from(proto_unwrap(src)));
}

}//namespace Fred::Registry::{anonymous}

template <>
Nsset::NssetHistoryInterval proto_unwrap(const Api::Nsset::NssetHistoryInterval& src)
{
    auto lower_limit = Nsset::NssetHistoryInterval::LowerLimit(Nsset::NssetHistoryInterval::NoLimit());
    if (src.has_lower_limit())
    {
        if (src.lower_limit().has_nsset_history_id())
        {
            lower_limit.value = unwrap_nsset_history_interval_history_id(src.lower_limit().nsset_history_id());
        }
        else if (src.lower_limit().has_timestamp())
        {
            lower_limit.value = unwrap_system_clock_nanoseconds(src.lower_limit().timestamp());
        }
    }
    auto upper_limit = Nsset::NssetHistoryInterval::UpperLimit(Nsset::NssetHistoryInterval::NoLimit());
    if (src.has_upper_limit())
    {
        if (src.upper_limit().has_nsset_history_id())
        {
            upper_limit.value = unwrap_nsset_history_interval_history_id(src.upper_limit().nsset_history_id());
        }
        else if (src.upper_limit().has_timestamp())
        {
            upper_limit.value = unwrap_system_clock_nanoseconds(src.upper_limit().timestamp());
        }
    }
    return Nsset::NssetHistoryInterval(lower_limit, upper_limit);
}

template <>
Nsset::SearchNssetRequest proto_unwrap(const Api::Nsset::SearchNssetRequest& src)
{
    Nsset::SearchNssetRequest dst;
    dst.query_values.reserve(src.query_values_size());
    for (const auto& value : src.query_values())
    {
        dst.query_values.push_back(value);
    }
    if (src.Limit_case() != Api::Nsset::SearchNssetRequest::LIMIT_NOT_SET)
    {
        dst.limit = src.limit();
    }
    else
    {
        dst.limit = boost::none;
    }
    for (const auto& item : src.searched_items())
    {
        if (item == Util::Into<std::string>::from(Nsset::NssetItem::nsset_handle))
        {
            dst.searched_items.insert(Nsset::NssetItem::nsset_handle);
        }
        else if (item == Util::Into<std::string>::from(Nsset::NssetItem::dns_hosts_fqdn))
        {
            dst.searched_items.insert(Nsset::NssetItem::dns_hosts_fqdn);
        }
        else if (item == Util::Into<std::string>::from(Nsset::NssetItem::dns_hosts_ip_addresses))
        {
            dst.searched_items.insert(Nsset::NssetItem::dns_hosts_ip_addresses);
        }
        else
        {
            LOGGER.warning("unknown item \"" + item + "\" in SearchNssetRequest");
        }
    }
    if (dst.searched_items.empty())
    {
        dst.searched_items = { Nsset::NssetItem::nsset_handle,
                               Nsset::NssetItem::dns_hosts_fqdn,
                               Nsset::NssetItem::dns_hosts_ip_addresses };
    }
    return dst;
}

template <>
Nsset::SearchNssetHistoryRequest proto_unwrap(const Api::Nsset::SearchNssetHistoryRequest& src)
{
    Nsset::SearchNssetHistoryRequest dst;
    dst.query_values.reserve(src.query_values_size());
    for (const auto& value : src.query_values())
    {
        dst.query_values.push_back(value);
    }
    if (src.Limit_case() != Api::Nsset::SearchNssetHistoryRequest::LIMIT_NOT_SET)
    {
        dst.max_number_of_nssets = src.max_number_of_nssets();
    }
    else
    {
        dst.max_number_of_nssets = boost::none;
    }
    if (src.has_data_valid_from())
    {
        dst.data_valid_from = unwrap_system_clock_nanoseconds(src.data_valid_from());
    }
    else
    {
        dst.data_valid_from = boost::none;
    }
    if (src.has_data_valid_to())
    {
        dst.data_valid_to = unwrap_system_clock_nanoseconds(src.data_valid_to());
    }
    else
    {
        dst.data_valid_to = boost::none;
    }
    for (const auto& item : src.searched_items())
    {
        if (item == Util::Into<std::string>::from(Nsset::NssetItem::nsset_handle))
        {
            dst.searched_items.insert(Nsset::NssetItem::nsset_handle);
        }
        else if (item == Util::Into<std::string>::from(Nsset::NssetItem::dns_hosts_fqdn))
        {
            dst.searched_items.insert(Nsset::NssetItem::dns_hosts_fqdn);
        }
        else if (item == Util::Into<std::string>::from(Nsset::NssetItem::dns_hosts_ip_addresses))
        {
            dst.searched_items.insert(Nsset::NssetItem::dns_hosts_ip_addresses);
        }
        else
        {
            LOGGER.warning("unknown item \"" + item + "\" in SearchNssetHistoryRequest");
        }
    }
    if (dst.searched_items.empty())
    {
        dst.searched_items = { Nsset::NssetItem::nsset_handle,
                               Nsset::NssetItem::dns_hosts_fqdn,
                               Nsset::NssetItem::dns_hosts_ip_addresses };
    }
    return dst;
}

template <>
Nsset::NssetDataHistoryRequest proto_unwrap(const Api::Nsset::NssetHistoryRequest& src)
{
    return Nsset::NssetDataHistoryRequest(
            proto_unwrap(src.nsset_id()),
            proto_unwrap(src.history()));
}

template <>
Nsset::NssetHandleHistoryRequest proto_unwrap(const Api::Nsset::NssetHandleHistoryRequest& src)
{
    return Nsset::NssetHandleHistoryRequest(proto_unwrap(src.nsset_handle()));
}

template <>
Nsset::NssetStateRequest proto_unwrap(const Api::Nsset::NssetStateRequest& src)
{
    Nsset::NssetStateRequest dst;
    dst.nsset_id = proto_unwrap(src.nsset_id());
    return dst;
}

template <>
Nsset::NssetStateHistoryRequest proto_unwrap(const Api::Nsset::NssetStateHistoryRequest& src)
{
    return Nsset::NssetStateHistoryRequest(proto_unwrap(src.nsset_id()), proto_unwrap(src.history()));
}

template <>
Domain::DomainId proto_unwrap(const Api::Domain::DomainId& src)
{
    return Util::make_strong<Domain::DomainId>(proto_unwrap(src.uuid()));
}

template <>
Domain::DomainHistoryId proto_unwrap(const Api::Domain::DomainHistoryId& src)
{
    return Util::make_strong<Domain::DomainHistoryId>(proto_unwrap(src.uuid()));
}

template <>
Domain::DomainInfoRequest proto_unwrap(const Api::Domain::DomainInfoRequest& src)
{
    Domain::DomainInfoRequest dst;
    dst.domain_id = proto_unwrap(src.domain_id());
    if (src.has_domain_history_id())
    {
        dst.domain_history_id = proto_unwrap(src.domain_history_id());
    }
    return dst;
}

template <>
Domain::Fqdn proto_unwrap(const Api::Domain::Fqdn& src)
{
    return Util::make_strong<Domain::Fqdn>(proto_unwrap(src.value()));
}

template <>
Domain::DomainIdRequest proto_unwrap(const Api::Domain::DomainIdRequest& src)
{
    Domain::DomainIdRequest dst;
    dst.fqdn = proto_unwrap(src.fqdn());
    return dst;
}

template <>
Domain::DomainStateFlagRequest proto_unwrap(const Api::Domain::DomainStateFlagRequest& src)
{
    return Util::make_strong<Domain::DomainStateFlagRequest>(proto_unwrap(src.uuid()));
}

namespace {

Domain::DomainHistoryInterval::HistoryId unwrap_domain_history_interval_history_id(const Api::Domain::DomainHistoryId& src)
{
    return Util::make_strong<Domain::DomainHistoryInterval::HistoryId>(
            Util::get_raw_value_from(proto_unwrap(src)));
}

Domain::BatchDeleteDomainsRequest::Domain proto_unwrap_domain(const Api::Domain::BatchDeleteDomainsRequest::Domain& src)
{
    Domain::BatchDeleteDomainsRequest::Domain dst{};
    dst.domain_id = proto_unwrap(src.domain_id());
    dst.domain_history_id = proto_unwrap(src.domain_history_id());
    return dst;
}

Domain::ManageDomainStateFlagsRequest::Domain proto_unwrap_domain(const Api::Domain::ManageDomainStateFlagsRequest::Domain& src)
{
    Domain::ManageDomainStateFlagsRequest::Domain dst{};
    dst.domain_id = proto_unwrap(src.domain_id());
    if (src.has_domain_history_id())
    {
        dst.domain_history_id = proto_unwrap(src.domain_history_id());
    }
    else
    {
        dst.domain_history_id = boost::none;
    }
    return dst;
}

Domain::ManageDomainStateFlagsRequest::Validity proto_unwrap_validity(const Api::Domain::ManageDomainStateFlagsRequest::Validity& src)
{
    Domain::ManageDomainStateFlagsRequest::Validity dst{};
    if (src.has_valid_from())
    {
        dst.valid_from = unwrap_system_clock_nanoseconds(src.valid_from());
    }
    else
    {
        dst.valid_from = boost::none;
    }
    if (src.has_valid_to())
    {
        dst.valid_to = unwrap_system_clock_nanoseconds(src.valid_to());
    }
    else
    {
        dst.valid_to = boost::none;
    }
    return dst;
}

Domain::ManageDomainStateFlagsRequest::ToCreate proto_unwrap_request_to_create(const Api::Domain::ManageDomainStateFlagsRequest::ToCreate& src)
{
    Domain::ManageDomainStateFlagsRequest::ToCreate dst{};
    std::for_each(begin(src.state_flags()), end(src.state_flags()), [&](auto&& state_flag)
            {
                dst.state_flags[state_flag.first] = proto_unwrap_validity(state_flag.second);
            });
    dst.domains.reserve(dst.domains.size() + src.domains().size());
    for (int idx = 0; idx < src.domains().size(); ++idx)
    {
        dst.domains.push_back(proto_unwrap_domain(src.domains(idx)));
    }
    return dst;
}

}//namespace Fred::Registry::{anonymous}

void proto_unwrap_back(const Api::Domain::BatchDeleteDomainsRequest& src, Domain::BatchDeleteDomainsRequest& dst)
{
    dst.domains.reserve(dst.domains.size() + src.domains().size());
    for (int idx = 0; idx < src.domains().size(); ++idx)
    {
        dst.domains.push_back(proto_unwrap_domain(src.domains(idx)));
    }
}

void proto_unwrap_back(const Api::Domain::ManageDomainStateFlagsRequest& src, Domain::ManageDomainStateFlagsRequest& dst)
{
    dst.state_flag_requests_to_create = proto_unwrap_request_to_create(src.state_flag_requests_to_create());
    dst.state_flag_requests_to_close.reserve(dst.state_flag_requests_to_close.size() + src.state_flag_requests_to_close().size());
    for (int idx = 0; idx < src.state_flag_requests_to_close().size(); ++idx)
    {
        dst.state_flag_requests_to_close.push_back(proto_unwrap(src.state_flag_requests_to_close(idx)));
    }
}

template <>
Domain::UpdateDomainsAdditionalNotifyInfoRequest::DomainAdditionalInfo proto_unwrap(const Api::Domain::UpdateDomainsDeleteAdditionalNotifyInfoRequest::DomainAdditionalInfo& src)
{
    Domain::UpdateDomainsAdditionalNotifyInfoRequest::DomainAdditionalInfo dst;
    dst.domain_id = proto_unwrap(src.domain_id());
    dst.event_time.lower_bound = unwrap_system_clock_nanoseconds(src.delete_from());
    dst.event_time.upper_bound = unwrap_system_clock_nanoseconds(src.delete_to());
    for (int idx = 0; idx < src.additional_contacts().size(); ++idx)
    {
        switch (src.additional_contacts(idx).Info_case())
        {
            case Api::Domain::DomainContactInfo::kEmail:
                dst.additional_emails.emplace_back(Util::make_strong<Domain::EmailAddress>(src.additional_contacts(idx).email()));
                break;
            case Api::Domain::DomainContactInfo::kTelephone:
                dst.additional_phones.emplace_back(Util::make_strong<Domain::PhoneNumber>(src.additional_contacts(idx).telephone()));
                break;
            case Api::Domain::DomainContactInfo::INFO_NOT_SET:
                break;
        }
    }
    return dst;
}

template <>
Domain::UpdateDomainsAdditionalNotifyInfoRequest::DomainAdditionalInfo proto_unwrap(const Api::Domain::UpdateDomainsOutzoneAdditionalNotifyInfoRequest::DomainAdditionalInfo& src)
{
    Domain::UpdateDomainsAdditionalNotifyInfoRequest::DomainAdditionalInfo dst;
    dst.domain_id = proto_unwrap(src.domain_id());
    dst.event_time.lower_bound = unwrap_system_clock_nanoseconds(src.outzone_from());
    dst.event_time.upper_bound = unwrap_system_clock_nanoseconds(src.outzone_to());
    for (int idx = 0; idx < src.additional_contacts().size(); ++idx)
    {
        switch (src.additional_contacts(idx).Info_case())
        {
            case Api::Domain::DomainContactInfo::kEmail:
                dst.additional_emails.emplace_back(Util::make_strong<Domain::EmailAddress>(src.additional_contacts(idx).email()));
                break;
            case Api::Domain::DomainContactInfo::kTelephone:
                dst.additional_phones.emplace_back(Util::make_strong<Domain::PhoneNumber>(src.additional_contacts(idx).telephone()));
                break;
            case Api::Domain::DomainContactInfo::INFO_NOT_SET:
                break;
        }
    }
    return dst;
}

void proto_unwrap_back(const Api::Domain::UpdateDomainsDeleteAdditionalNotifyInfoRequest& src, Domain::UpdateDomainsAdditionalNotifyInfoRequest& dst)
{
    dst.event = Domain::LifecycleEvent::domain_delete;
    dst.domains_additional_info.reserve(dst.domains_additional_info.size() + src.domains_additional_info().size());
    for (int idx = 0; idx < src.domains_additional_info().size(); ++idx)
    {
        dst.domains_additional_info.push_back(proto_unwrap(src.domains_additional_info(idx)));
    }
}

void proto_unwrap_back(const Api::Domain::UpdateDomainsOutzoneAdditionalNotifyInfoRequest& src, Domain::UpdateDomainsAdditionalNotifyInfoRequest& dst)
{
    dst.event = Domain::LifecycleEvent::domain_outzone;
    dst.domains_additional_info.reserve(dst.domains_additional_info.size() + src.domains_additional_info().size());
    for (int idx = 0; idx < src.domains_additional_info().size(); ++idx)
    {
        dst.domains_additional_info.push_back(proto_unwrap(src.domains_additional_info(idx)));
    }
}

template <>
Domain::DomainHistoryInterval proto_unwrap(const Api::Domain::DomainHistoryInterval& src)
{
    auto lower_limit = Domain::DomainHistoryInterval::LowerLimit(Domain::DomainHistoryInterval::NoLimit());
    if (src.has_lower_limit())
    {
        if (src.lower_limit().has_domain_history_id())
        {
            lower_limit.value = unwrap_domain_history_interval_history_id(src.lower_limit().domain_history_id());
        }
        else if (src.lower_limit().has_timestamp())
        {
            lower_limit.value = unwrap_system_clock_nanoseconds(src.lower_limit().timestamp());
        }
    }
    auto upper_limit = Domain::DomainHistoryInterval::UpperLimit(Domain::DomainHistoryInterval::NoLimit());
    if (src.has_upper_limit())
    {
        if (src.upper_limit().has_domain_history_id())
        {
            upper_limit.value = unwrap_domain_history_interval_history_id(src.upper_limit().domain_history_id());
        }
        else if (src.upper_limit().has_timestamp())
        {
            upper_limit.value = unwrap_system_clock_nanoseconds(src.upper_limit().timestamp());
        }
    }
    return Domain::DomainHistoryInterval(lower_limit, upper_limit);
}

template <>
Domain::SearchDomainRequest proto_unwrap(const Api::Domain::SearchDomainRequest& src)
{
    Domain::SearchDomainRequest dst;
    for (const auto& item : src.query_values())
    {
        dst.query_values.push_back(item);
    }

    if (src.Limit_case() != Api::Domain::SearchDomainRequest::LIMIT_NOT_SET)
    {
        dst.limit = src.limit();
    }
    else
    {
        dst.limit = boost::none;
    }
    for (const auto& item : src.searched_items())
    {
        if (item == Util::Into<std::string>::from(Domain::DomainItem::fqdn))
        {
            dst.searched_items.insert(Domain::DomainItem::fqdn);
        }
        else
        {
            LOGGER.warning("unknown item \"" + item + "\" in SearchDomainRequest");
        }
    }
    if (dst.searched_items.empty())
    {
        dst.searched_items.insert(Domain::DomainItem::fqdn);
    }
    return dst;
}

template <>
Domain::SearchDomainHistoryRequest proto_unwrap(const Api::Domain::SearchDomainHistoryRequest& src)
{
    Domain::SearchDomainHistoryRequest dst;
    dst.query_values.reserve(src.query_values_size());
    for (const auto& value : src.query_values())
    {
        dst.query_values.push_back(value);
    }
    if (src.Limit_case() != Api::Domain::SearchDomainHistoryRequest::LIMIT_NOT_SET)
    {
        dst.max_number_of_domains = src.max_number_of_domains();
    }
    else
    {
        dst.max_number_of_domains = boost::none;
    }
    if (src.has_data_valid_from())
    {
        dst.data_valid_from = unwrap_system_clock_nanoseconds(src.data_valid_from());
    }
    else
    {
        dst.data_valid_from = boost::none;
    }
    if (src.has_data_valid_to())
    {
        dst.data_valid_to = unwrap_system_clock_nanoseconds(src.data_valid_to());
    }
    else
    {
        dst.data_valid_to = boost::none;
    }
    for (const auto& item : src.searched_items())
    {
        if (item == Util::Into<std::string>::from(Domain::DomainItem::fqdn))
        {
            dst.searched_items.insert(Domain::DomainItem::fqdn);
        }
        else
        {
            LOGGER.warning("unknown item \"" + item + "\" in SearchDomainHistoryRequest");
        }
    }
    if (dst.searched_items.empty())
    {
        dst.searched_items.insert(Domain::DomainItem::fqdn);
    }
    return dst;
}

template <>
Domain::DomainDataHistoryRequest proto_unwrap(const Api::Domain::DomainHistoryRequest& src)
{
    return Domain::DomainDataHistoryRequest(
            proto_unwrap(src.domain_id()),
            proto_unwrap(src.history()));
}

template <>
Domain::FqdnHistoryRequest proto_unwrap(const Api::Domain::FqdnHistoryRequest& src)
{
    return Domain::FqdnHistoryRequest(proto_unwrap(src.fqdn()));
}

template <>
Domain::DomainStateRequest proto_unwrap(const Api::Domain::DomainStateRequest& src)
{
    Domain::DomainStateRequest dst;
    dst.domain_id = proto_unwrap(src.domain_id());
    return dst;
}

template <>
Domain::DomainStateHistoryRequest proto_unwrap(const Api::Domain::DomainStateHistoryRequest& src)
{
    return Domain::DomainStateHistoryRequest(proto_unwrap(src.domain_id()), proto_unwrap(src.history()));
}

template <>
PaginationRequest proto_unwrap(const Api::PaginationRequest& src)
{
    PaginationRequest dst;
    dst.page_size = src.page_size();
    dst.page_token = proto_unwrap(src.page_token());
    return dst;
}

template <>
Domain::GetDomainsByContactRequest proto_unwrap(const Api::Domain::DomainsByContactRequest& src)
{
    Domain::GetDomainsByContactRequest dst;
    dst.contact_id = proto_unwrap(src.contact_id());
    if (src.has_pagination())
    {
        dst.pagination = proto_unwrap(src.pagination());
    }
    dst.order_by = proto_unwrap(src.order_by());
    dst.fqdn_filter = proto_unwrap(src.fqdn_filter());
    dst.include_holder = src.include_holder();
    dst.include_admin_contact = src.include_admin_contact();
    dst.include_keyset_tech_contact = src.include_keyset_tech_contact();
    dst.include_nsset_tech_contact = src.include_nsset_tech_contact();
    return dst;
}

template <>
Domain::GetDomainsByHolderRequest proto_unwrap(const Api::Domain::DomainsByHolderRequest& src)
{
    Domain::GetDomainsByHolderRequest dst;
    dst.domain_holder_id = proto_unwrap(src.domain_holder_id());
    if (src.has_domain_holder_history_id())
    {
        dst.domain_holder_history_id = proto_unwrap(src.domain_holder_history_id());
    }
    return dst;
}

template <>
Domain::GetDomainsNotifyInfoRequest proto_unwrap(const Api::Domain::GetDomainsOutzoneNotifyInfoRequest& src)
{
    Domain::GetDomainsNotifyInfoRequest dst;
    dst.event_time.lower_bound = unwrap_system_clock_nanoseconds(src.outzone_from());
    dst.event_time.upper_bound = unwrap_system_clock_nanoseconds(src.outzone_to());
    if (src.has_pagination())
    {
        dst.pagination = proto_unwrap(src.pagination());
    }
    dst.event = Domain::LifecycleEvent::domain_outzone;
    return dst;
}

template <>
Domain::GetDomainsNotifyInfoRequest proto_unwrap(const Api::Domain::GetDomainsDeleteNotifyInfoRequest& src)
{
    Domain::GetDomainsNotifyInfoRequest dst;
    dst.event_time.lower_bound = unwrap_system_clock_nanoseconds(src.delete_from());
    dst.event_time.upper_bound = unwrap_system_clock_nanoseconds(src.delete_to());
    if (src.has_pagination())
    {
        dst.pagination = proto_unwrap(src.pagination());
    }
    dst.event = Domain::LifecycleEvent::domain_delete;
    return dst;
}


template <>
struct ProtoTypeTraits<Api::Registrar::RegistrarHandle>
{
    using RegistryType = Registrar::RegistrarHandle;
};

template <>
Registrar::RegistrarHandle proto_unwrap(const Api::Registrar::RegistrarHandle& src)
{
    return Util::make_strong<Registrar::RegistrarHandle>(proto_unwrap(src.value()));
}

template <>
Registrar::RegistrarCreditRequest proto_unwrap(const Api::Registrar::RegistrarCreditRequest& src)
{
    Registrar::RegistrarCreditRequest dst;
    dst.registrar_handle = proto_unwrap(src.registrar_handle());
    return dst;
}

template <>
Registrar::RegistrarInfoRequest proto_unwrap(const Api::Registrar::RegistrarInfoRequest& src)
{
    Registrar::RegistrarInfoRequest dst;
    dst.registrar_handle = proto_unwrap(src.registrar_handle());
    return dst;
}

template <>
Registrar::RegistrarZoneAccessRequest proto_unwrap(const Api::Registrar::RegistrarZoneAccessRequest& src)
{
    Registrar::RegistrarZoneAccessRequest dst;
    dst.registrar_handle = proto_unwrap(src.registrar_handle());
    return dst;
}

}//namespace Fred::Registry
}//namespace Fred
