/*
 * Copyright (C) 2018-2022  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/proto_wrap.hh"
#include "src/util/into.hh"

#include <boost/lexical_cast.hpp>
#include <boost/uuid/string_generator.hpp>
#include <boost/uuid/uuid_io.hpp>

namespace Fred {
namespace Registry {

namespace {

template <typename N>
void proto_wrap_uuid(const Uuid<N>& src, Api::Uuid* dst)
{
    dst->set_value(boost::lexical_cast<std::string>(get_raw_value_from(src)));
}

}//namespace Fred::Registry::{anonymous}

template <>
void proto_wrap(const LogEntryId& src, Api::LogEntryId* dst)
{
    dst->set_value(get_raw_value_from(src));
}

template <>
struct RegistryTypeTraits<Registrar::RegistrarHandle>
{
    using ProtoType = Api::Registrar::RegistrarHandle;
};

template <>
void proto_wrap(const Registrar::RegistrarHandle& src, Api::Registrar::RegistrarHandle* dst)
{
    dst->set_value(get_raw_value_from(src));
}

template <>
struct RegistryTypeTraits<Keyset::KeysetId>
{
    using ProtoType = Api::Keyset::KeysetId;
};

template <>
void proto_wrap(const Keyset::KeysetId& src, Api::Keyset::KeysetId* dst)
{
    proto_wrap_uuid(src, dst->mutable_uuid());
}

template <>
struct RegistryTypeTraits<Nsset::NssetId>
{
    using ProtoType = Api::Nsset::NssetId;
};

template <>
void proto_wrap(const Nsset::NssetId& src, Api::Nsset::NssetId* dst)
{
    proto_wrap_uuid(src, dst->mutable_uuid());
}

namespace {

void proto_wrap(const Contact::ContactHandle& src, Api::Contact::ContactHandle* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Contact::ContactId& src, Api::Contact::ContactId* dst)
{
    proto_wrap_uuid(src, dst->mutable_uuid());
}

void proto_wrap(const Contact::ContactHistoryId& src, Api::Contact::ContactHistoryId* dst)
{
    proto_wrap_uuid(src, dst->mutable_uuid());
}

void proto_wrap(const Contact::PostalCode& src, Api::PostalCode* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Contact::CountryCode& src, Api::CountryCode* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Contact::PlaceAddress& src, Api::PlaceAddress* dst)
{
    for (const auto& street : src.street)
    {
        dst->add_street(street);
    }
    dst->set_city(src.city);
    dst->set_state_or_province(src.state_or_province);
    proto_wrap(src.postal_code, dst->mutable_postal_code());
    proto_wrap(src.country_code, dst->mutable_country_code());
}

void proto_wrap(const Contact::PrivacyControlled<Contact::PlaceAddress>& src, Api::PlaceAddress* dst)
{
    proto_wrap(get_privacy_controlled_data(src), dst);
}

void proto_wrap(const Contact::PhoneNumber& src, Api::PhoneNumber* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Contact::EmailAddress& src, Api::EmailAddress* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Contact::VatIdentificationNumber& src, Api::VatIdentificationNumber* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Contact::NationalIdentityNumber& src, Api::Contact::NationalIdentityNumber* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Contact::NationalIdentityCard& src, Api::Contact::NationalIdentityCard* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Contact::PassportNumber& src, Api::Contact::PassportNumber* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Contact::CompanyRegistrationNumber& src, Api::CompanyRegistrationNumber* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Contact::SocialSecurityNumber& src, Api::Contact::SocialSecurityNumber* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Contact::Birthdate& src, Api::Contact::Birthdate* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Contact::ContactAdditionalIdentifier& src, Api::Contact::ContactAdditionalIdentifier* dst)
{
    class Visitor : public boost::static_visitor<void>
    {
    public:
        explicit Visitor(Api::Contact::ContactAdditionalIdentifier* dst) : dst_(dst) { }
        void operator()(const Contact::NationalIdentityNumber& src) const
        {
            proto_wrap(src, dst_->mutable_national_identity_number());
        }
        void operator()(const Contact::NationalIdentityCard& src) const
        {
            proto_wrap(src, dst_->mutable_national_identity_card());
        }
        void operator()(const Contact::PassportNumber& src) const
        {
            proto_wrap(src, dst_->mutable_passport_number());
        }
        void operator()(const Contact::CompanyRegistrationNumber& src) const
        {
            proto_wrap(src, dst_->mutable_company_registration_number());
        }
        void operator()(const Contact::SocialSecurityNumber& src) const
        {
            proto_wrap(src, dst_->mutable_social_security_number());
        }
        void operator()(const Contact::Birthdate& src) const
        {
            proto_wrap(src, dst_->mutable_birthdate());
        }
    private:
        Api::Contact::ContactAdditionalIdentifier* dst_;
    };
    boost::apply_visitor(Visitor(dst), src);
}

void proto_wrap(const Contact::ContactAddress& src, Api::Contact::ContactAddress* dst)
{
    dst->set_company(src.company);
    for (const auto& street : src.street)
    {
        dst->add_street(street);
    }
    dst->set_city(src.city);
    dst->set_state_or_province(src.state_or_province);
    proto_wrap(src.postal_code, dst->mutable_postal_code());
    proto_wrap(src.country_code, dst->mutable_country_code());
}

void proto_wrap(const Contact::WarningLetter& src, Api::Contact::WarningLetter* dst)
{
    switch (src.preference)
    {
        case Contact::WarningLetter::SendingPreference::to_send:
            dst->set_preference(Api::Contact::WarningLetter::SendingPreference::to_send);
            return;
        case Contact::WarningLetter::SendingPreference::not_to_send:
            dst->set_preference(Api::Contact::WarningLetter::SendingPreference::not_to_send);
            return;
        case Contact::WarningLetter::SendingPreference::not_specified:
            dst->set_preference(Api::Contact::WarningLetter::SendingPreference::not_specified);
            return;
    }
    throw std::runtime_error("unknown warning letter preference");
}

void proto_wrap(const Contact::AuthInfo& src, Registry::Api::AuthInfo* dst)
{
    dst->set_password(src.password);
}

template <typename R>
void proto_wrap(const std::chrono::time_point<std::chrono::system_clock, R>& src, ::google::protobuf::Timestamp* dst)
{
    const auto time_in_nanos =
            std::chrono::duration_cast<std::chrono::nanoseconds>(src.time_since_epoch()).count();
    dst->set_seconds(time_in_nanos / 1'000'000'000);
    dst->set_nanos(time_in_nanos % 1'000'000'000);
}

void proto_wrap(
        const Object::ObjectEvents::EventData& src,
        Registry::Api::ObjectEvents::EventData* dst)
{
    if (src.at != boost::none)
    {
        proto_wrap(*src.at, dst->mutable_timestamp());
    }
    else
    {
        dst->clear_timestamp();
    }
    if (src.by_registrar != boost::none)
    {
        Registry::proto_wrap(*src.by_registrar, dst->mutable_registrar_handle());
    }
    else
    {
        dst->clear_registrar_handle();
    }
}

void proto_wrap(const Object::ObjectEvents& src, Registry::Api::ObjectEvents* dst)
{
    proto_wrap(src.registered, dst->mutable_registered());
    if (src.transferred != boost::none)
    {
        proto_wrap(*src.transferred, dst->mutable_transferred());
    }
    else
    {
        dst->clear_transferred();
    }
    if (src.updated != boost::none)
    {
        proto_wrap(*src.updated, dst->mutable_updated());
    }
    else
    {
        dst->clear_updated();
    }
    if (src.unregistered != boost::none)
    {
        proto_wrap(*src.unregistered, dst->mutable_unregistered());
    }
    else
    {
        dst->clear_unregistered();
    }
}

template <typename T>
void proto_wrap_contact_discloseflags(const std::string& name, bool is_available_publicly, T& dst)
{
    dst[name] = is_available_publicly;
}

template <typename T>
void proto_wrap_publishability(const Contact::ContactInfoReply::Data& src, T* dst)
{
    proto_wrap_contact_discloseflags("name", is_publicly_available(src.name), *dst);
    proto_wrap_contact_discloseflags("organization", is_publicly_available(src.organization), *dst);
    proto_wrap_contact_discloseflags("place", is_publicly_available(src.place), *dst);
    proto_wrap_contact_discloseflags("telephone", is_publicly_available(src.telephone), *dst);
    proto_wrap_contact_discloseflags("fax", is_publicly_available(src.fax), *dst);
    proto_wrap_contact_discloseflags("emails", is_publicly_available(src.emails), *dst);
    proto_wrap_contact_discloseflags("notify_emails", is_publicly_available(src.notify_emails), *dst);
    proto_wrap_contact_discloseflags("vat_identification_number", is_publicly_available(src.vat_identification_number), *dst);
    proto_wrap_contact_discloseflags("additional_identifier", is_publicly_available(src.additional_identifier), *dst);
}

template <typename T>
void proto_wrap_emails(const std::vector<Contact::EmailAddress>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& email : src)
    {
        proto_wrap(email, dst->Add());
    }
}

void proto_wrap(const Contact::ContactInfoReply::Data& src, Api::Contact::ContactInfoReply::Data* dst)
{
    proto_wrap(src.contact_handle, dst->mutable_contact_handle());
    proto_wrap(src.contact_id, dst->mutable_contact_id());
    proto_wrap(src.contact_history_id, dst->mutable_contact_history_id());
    dst->set_name(get_privacy_controlled_data(src.name));
    dst->set_organization(get_privacy_controlled_data(src.organization));
    proto_wrap(src.place, dst->mutable_place());
    if (does_present_privacy_controlled_data(src.telephone))
    {
        proto_wrap(*get_privacy_controlled_data(src.telephone), dst->mutable_telephone());
    }
    if (does_present_privacy_controlled_data(src.fax))
    {
        proto_wrap(*get_privacy_controlled_data(src.fax), dst->mutable_fax());
    }
    proto_wrap_emails(get_privacy_controlled_data(src.emails), dst->mutable_emails());
    proto_wrap_emails(get_privacy_controlled_data(src.notify_emails), dst->mutable_notify_emails());
    if (does_present_privacy_controlled_data(src.vat_identification_number))
    {
        proto_wrap(*get_privacy_controlled_data(src.vat_identification_number), dst->mutable_vat_identification_number());
    }
    if (does_present_privacy_controlled_data(src.additional_identifier))
    {
        proto_wrap(*get_privacy_controlled_data(src.additional_identifier), dst->mutable_additional_identifier());
    }
    proto_wrap_publishability(src, dst->mutable_publish());
    if (src.mailing_address != boost::none)
    {
        proto_wrap(*src.mailing_address, dst->mutable_mailing_address());
    }
    if (src.billing_address != boost::none)
    {
        proto_wrap(*src.billing_address, dst->mutable_billing_address());
    }
    if (src.shipping_address[0] != boost::none)
    {
        proto_wrap(*src.shipping_address[0], dst->mutable_shipping_address1());
    }
    if (src.shipping_address[1] != boost::none)
    {
        proto_wrap(*src.shipping_address[1], dst->mutable_shipping_address2());
    }
    if (src.shipping_address[2] != boost::none)
    {
        proto_wrap(*src.shipping_address[2], dst->mutable_shipping_address3());
    }
    proto_wrap(src.warning_letter, dst->mutable_warning_letter());
    proto_wrap(src.auth_info, dst->mutable_auth_info());
    proto_wrap(src.representative_events, dst->mutable_events());
    Registry::proto_wrap(src.sponsoring_registrar, dst->mutable_sponsoring_registrar());
}

void proto_wrap(const Contact::ContactIdReply::Data& src, Api::Contact::ContactIdReply::Data* dst)
{
    proto_wrap(src.contact_id, dst->mutable_contact_id());
    proto_wrap(src.contact_history_id, dst->mutable_contact_history_id());
}

template <typename T>
void proto_wrap_contact_state_flags(const Contact::ContactStateFlagsInfo& src, T* dst)
{
    for (const auto& flag : src)
    {
        Registry::Api::StateFlag::Traits traits;
        switch (flag.how_to_set)
        {
            case Contact::StateFlagInfo::Manipulation::automatic:
                traits.set_how_to_set(Registry::Api::StateFlag::Manipulation::automatic);
                break;
            case Contact::StateFlagInfo::Manipulation::manual:
                traits.set_how_to_set(Registry::Api::StateFlag::Manipulation::manual);
                break;
        }
        switch (flag.visibility)
        {
            case Contact::StateFlagInfo::Visibility::external:
                traits.set_visibility(Registry::Api::StateFlag::Visibility::external);
                break;
            case Contact::StateFlagInfo::Visibility::internal:
                traits.set_visibility(Registry::Api::StateFlag::Visibility::internal);
                break;
        }
        (*dst)[flag.name] = traits;
    }
}

template <typename T>
void proto_wrap_contact_state(const Contact::ContactState& src, T* dst)
{
    for (const auto& name_presence : src.flag_presents)
    {
        (*dst)[name_presence.first] = name_presence.second;
    }
}

void proto_wrap(
        const Contact::ContactStateHistory::Record& record,
        const std::vector<std::string>& flags_names,
        Api::Contact::ContactStateHistory::Record* dst)
{
    if (record.presents.size() != flags_names.size())
    {
        throw std::runtime_error("number of flags must be equal to the number of flags names");
    }
    proto_wrap(record.valid_from, dst->mutable_valid_from());
    for (std::size_t idx = 0; idx < record.presents.size(); ++idx)
    {
        (*(dst->mutable_flags()))[flags_names[idx]] = record.presents[idx];
    }
}

void proto_wrap(const Contact::ContactStateHistory& src, Api::Contact::ContactStateHistory* dst)
{
    dst->mutable_timeline()->Reserve(src.timeline.size());
    for (const auto& record : src.timeline)
    {
        proto_wrap(record, src.flags_names, dst->mutable_timeline()->Add());
    }
    if (src.valid_to == boost::none)
    {
        dst->clear_valid_to();
    }
    else
    {
        proto_wrap(*src.valid_to, dst->mutable_valid_to());
    }
}

void proto_wrap(const Contact::ContactDataHistory::Record& src, Api::Contact::ContactHistoryRecord* dst)
{
    proto_wrap(src.contact_history_id, dst->mutable_contact_history_id());
    proto_wrap(src.valid_from, dst->mutable_valid_from());
    proto_wrap(src.log_entry_id, dst->mutable_log_entry_id());
}

template <typename T>
void proto_wrap_vector_of_contact_data_history_record(const std::vector<Contact::ContactDataHistory::Record>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Contact::ContactDataHistory& src, Api::Contact::ContactHistoryReply::Data* dst)
{
    proto_wrap(src.contact_id, dst->mutable_contact_id());
    proto_wrap_vector_of_contact_data_history_record(src.timeline, dst->mutable_timeline());
    if (src.valid_to != boost::none)
    {
        proto_wrap(*src.valid_to, dst->mutable_valid_to());
    }
    else
    {
        dst->clear_valid_to();
    }
}

void proto_wrap(const Contact::ContactDataHistoryReply::Data& src, Api::Contact::ContactHistoryReply::Data* dst)
{
    proto_wrap(src.history, dst);
}

void proto_wrap(const Contact::ContactLifetime::TimeSpec& src, Api::Contact::ContactLifetime::TimeSpec* dst)
{
    proto_wrap(src.contact_history_id, dst->mutable_contact_history_id());
    proto_wrap(src.timestamp, dst->mutable_timestamp());
}

void proto_wrap(const Contact::ContactLifetime& src, Api::Contact::ContactLifetime* dst)
{
    proto_wrap(src.contact_id, dst->mutable_contact_id());
    proto_wrap(src.begin, dst->mutable_registered_from());
    if (src.end != boost::none)
    {
        proto_wrap(*src.end, dst->mutable_registered_to());
    }
    else
    {
        dst->clear_registered_to();
    }
}

template <typename T>
void proto_wrap_vector_of_contact_lifetime(const std::vector<Contact::ContactLifetime>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Contact::ContactHandleHistoryReply::Data& src, Api::Contact::ContactHandleHistoryReply::Data* dst)
{
    proto_wrap(src.contact_handle, dst->mutable_contact_handle());
    proto_wrap_vector_of_contact_lifetime(src.timeline, dst->mutable_timeline());
}

template <typename T>
void proto_wrap_set_of_string(const std::set<std::string>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& str : src)
    {
        *(dst->Add()) = str;
    }
}

void proto_wrap(const Contact::SearchContactReply::Data::Result& src, Api::Contact::SearchContactReply::Data::Result* dst)
{
    proto_wrap(src.contact_id, dst->mutable_object_id());
    proto_wrap_set_of_string(src.matched_items, dst->mutable_matched_items());
}

template <typename T>
void proto_wrap_vector_of_search_contact_result(const std::vector<Contact::SearchContactReply::Data::Result>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Contact::SearchContactReply::Data::FuzzyValue& src,
                Api::Contact::SearchContactReply::Data::FuzzyValue* dst)
{
    dst->set_lower_estimate(src.lower_estimate);
    dst->set_upper_estimate(src.upper_estimate);
}

template <typename T>
void proto_wrap_set_of_contact_items(const std::set<Contact::ContactItem>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto item : src)
    {
        *(dst->Add()) = Util::Into<std::string>::from(item);
    }
}

void proto_wrap(const Contact::SearchContactReply::Data& src, Api::Contact::SearchContactReply::Data* dst)
{
    proto_wrap_vector_of_search_contact_result(src.most_similar_contacts, dst->mutable_results());
    proto_wrap(src.result_count, dst->mutable_estimated_total());
    proto_wrap_set_of_contact_items(src.searched_items, dst->mutable_searched_items());
}

template <typename T>
void proto_wrap_vector_of_contact_history_id(
        const std::vector<Contact::ContactHistoryId>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Contact::SearchContactHistoryReply::Data::Result::HistoryPeriod& src,
                Api::Contact::SearchContactHistoryReply::Data::Result::HistoryPeriod* dst)
{
    proto_wrap_set_of_string(src.matched_items, dst->mutable_matched_items());
    proto_wrap(src.valid_from, dst->mutable_valid_from());
    if (src.valid_to == boost::none)
    {
        dst->clear_valid_to();
    }
    else
    {
        proto_wrap(*src.valid_to, dst->mutable_valid_to());
    }
    proto_wrap_vector_of_contact_history_id(src.contact_history_ids, dst->mutable_contact_history_ids());
}

template <typename T>
void proto_wrap_vector_of_search_contact_history_result_histories(
    const std::vector<Contact::SearchContactHistoryReply::Data::Result::HistoryPeriod>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Contact::SearchContactHistoryReply::Data::Result& src,
                Api::Contact::SearchContactHistoryReply::Data::Result* dst)
{
    proto_wrap(src.object_id, dst->mutable_object_id());
    proto_wrap_vector_of_search_contact_history_result_histories(src.histories, dst->mutable_histories());
}

template <typename T>
void proto_wrap_vector_of_search_contact_history_result(const std::vector<Contact::SearchContactHistoryReply::Data::Result>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Contact::SearchContactHistoryReply::Data::FuzzyValue& src,
                Api::Contact::SearchContactHistoryReply::Data::FuzzyValue* dst)
{
    dst->set_lower_estimate(src.lower_estimate);
    dst->set_upper_estimate(src.upper_estimate);
}

void proto_wrap(const Contact::SearchContactHistoryReply::Data& src, Api::Contact::SearchContactHistoryReply::Data* dst)
{
    proto_wrap_vector_of_search_contact_history_result(src.results, dst->mutable_results());
    proto_wrap(src.result_count, dst->mutable_estimated_total());
    proto_wrap_set_of_contact_items(src.searched_items, dst->mutable_searched_items());
}

void proto_wrap(const Contact::ContactStateFlagsInfo& src, Api::Contact::ContactStateFlagsReply::Data* dst)
{
    proto_wrap_contact_state_flags(src, dst->mutable_traits());
}

}//namespace Fred::Registry::{anonymous}

template<>
void proto_wrap(const Contact::ContactInfoReply::Data& src, Api::Contact::ContactInfoReply* dst)
{
    dst->clear_exception();
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Contact::ContactInfoReply::Exception::ContactDoesNotExist&, Api::Contact::ContactInfoReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_contact_does_not_exist();
}

template<>
void proto_wrap(const Contact::ContactIdReply::Data& src, Api::Contact::ContactIdReply* dst)
{
    dst->clear_exception();
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Contact::ContactIdReply::Exception::ContactDoesNotExist&, Api::Contact::ContactIdReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_contact_does_not_exist();
}

template<>
void proto_wrap(const Contact::ContactStateFlagsInfo& src, Api::Contact::ContactStateFlagsReply* dst)
{
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Contact::ContactDataHistoryReply::Data& src, Api::Contact::ContactHistoryReply* dst)
{
    dst->clear_data();
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Contact::ContactDataHistoryReply::Exception::ContactDoesNotExist&,
                Api::Contact::ContactHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_contact_does_not_exist();
}

template<>
void proto_wrap(const Contact::ContactDataHistoryReply::Exception::InvalidHistoryInterval&,
                Api::Contact::ContactHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_invalid_history_interval();
}

template<>
void proto_wrap(const Contact::ContactHandleHistoryReply::Data& src, Api::Contact::ContactHandleHistoryReply* dst)
{
    dst->clear_data();
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Contact::ContactStateReply::Data& src, Api::Contact::ContactStateReply* dst)
{
    proto_wrap(src.contact_id, dst->mutable_data()->mutable_contact_id());
    proto_wrap_contact_state(src.state, dst->mutable_data()->mutable_state()->mutable_flags());
}

template<>
void proto_wrap(const Contact::ContactStateReply::Exception::ContactDoesNotExist&, Api::Contact::ContactStateReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_contact_does_not_exist();
}

template<>
void proto_wrap(const Contact::ContactStateHistoryReply::Data& src, Api::Contact::ContactStateHistoryReply* dst)
{
    proto_wrap(src.contact_id, dst->mutable_data()->mutable_contact_id());
    proto_wrap(src.history, dst->mutable_data()->mutable_history());
}

template<>
void proto_wrap(const Contact::ContactStateHistoryReply::Exception::ContactDoesNotExist&,
                Api::Contact::ContactStateHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_contact_does_not_exist();
}

template<>
void proto_wrap(const Contact::ContactStateHistoryReply::Exception::InvalidHistoryInterval&,
                Api::Contact::ContactStateHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_invalid_history_interval();
}

template<>
void proto_wrap(const Contact::UpdateContactReply::Data& src, Api::Contact::UpdateContactReply* dst)
{
    proto_wrap(src.contact_id, dst->mutable_data()->mutable_contact_id());
    proto_wrap(src.contact_history_id, dst->mutable_data()->mutable_contact_history_id());
    proto_wrap(src.contact_handle, dst->mutable_data()->mutable_contact_handle());
    proto_wrap(src.events, dst->mutable_data()->mutable_events());
}

template<>
void proto_wrap(const Contact::UpdateContactReply::Exception::ContactDoesNotExist&, Api::Contact::UpdateContactReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_contact_does_not_exist();
}

template<>
void proto_wrap(const Contact::UpdateContactReply::Exception::NotCurrentVersion&, Api::Contact::UpdateContactReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_not_current_version();
}

template<>
void proto_wrap(const Contact::SearchContactReply::Data& src, Api::Contact::SearchContactReply* dst)
{
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Contact::SearchContactHistoryReply::Data& src, Api::Contact::SearchContactHistoryReply* dst)
{
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Contact::SearchContactHistoryReply::Exception::ChronologyViolation&, Api::Contact::SearchContactHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_chronology_violation();
}

namespace {

void proto_wrap(const Domain::Fqdn& src, Api::Domain::Fqdn* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Domain::DomainId& src, Api::Domain::DomainId* dst)
{
    proto_wrap_uuid(src, dst->mutable_uuid());
}

void proto_wrap(const Domain::DomainStateFlagRequest& src, Api::Domain::DomainStateFlagRequest* dst)
{
    proto_wrap_uuid(src, dst->mutable_uuid());
}

void proto_wrap(const Domain::DomainHistoryId& src, Api::Domain::DomainHistoryId* dst)
{
    proto_wrap_uuid(src, dst->mutable_uuid());
}

template <typename T>
void proto_wrap_vector_of_contacts(const std::vector<Contact::ContactId>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Domain::ExpirationTimestamp& src, ::google::protobuf::Timestamp* dst)
{
    proto_wrap(src.timestamp, dst);
}

void proto_wrap(const Domain::AuthInfo& src, Registry::Api::AuthInfo* dst)
{
    dst->set_password(src.password);
}

void proto_wrap(const Domain::DomainInfoReply::Data& src, Api::Domain::DomainInfoReply::Data* dst)
{
    proto_wrap(src.fqdn, dst->mutable_fqdn());
    proto_wrap(src.domain_id, dst->mutable_domain_id());
    proto_wrap(src.domain_history_id, dst->mutable_domain_history_id());
    if (src.keyset != boost::none)
    {
        Registry::proto_wrap(*src.keyset, dst->mutable_keyset());
    }
    else
    {
        dst->clear_keyset();
    }
    if (src.nsset != boost::none)
    {
        Registry::proto_wrap(*src.nsset, dst->mutable_nsset());
    }
    else
    {
        dst->clear_nsset();
    }
    proto_wrap(src.registrant, dst->mutable_registrant());
    Registry::proto_wrap(src.sponsoring_registrar, dst->mutable_sponsoring_registrar());
    proto_wrap_vector_of_contacts(src.administrative_contacts, dst->mutable_administrative_contacts());
    proto_wrap(src.expires_at, dst->mutable_expires_at());
    proto_wrap(src.auth_info, dst->mutable_auth_info());
    proto_wrap(src.representative_events, dst->mutable_events());
}

void proto_wrap(const Domain::DomainIdReply::Data& src, Api::Domain::DomainIdReply::Data* dst)
{
    proto_wrap(src.domain_id, dst->mutable_domain_id());
    proto_wrap(src.domain_history_id, dst->mutable_domain_history_id());
}

template <typename T>
void proto_wrap_domain_state_flags(const Domain::DomainStateFlagsInfo& src, T* dst)
{
    for (const auto& flag : src)
    {
        Registry::Api::StateFlag::Traits traits;
        switch (flag.how_to_set)
        {
            case Domain::StateFlagInfo::Manipulation::automatic:
                traits.set_how_to_set(Registry::Api::StateFlag::Manipulation::automatic);
                break;
            case Domain::StateFlagInfo::Manipulation::manual:
                traits.set_how_to_set(Registry::Api::StateFlag::Manipulation::manual);
                break;
        }
        switch (flag.visibility)
        {
            case Domain::StateFlagInfo::Visibility::external:
                traits.set_visibility(Registry::Api::StateFlag::Visibility::external);
                break;
            case Domain::StateFlagInfo::Visibility::internal:
                traits.set_visibility(Registry::Api::StateFlag::Visibility::internal);
                break;
        }
        (*dst)[flag.name] = traits;
    }
}

template <typename T>
void proto_wrap_domain_state(const Domain::DomainState& src, T* dst)
{
    for (const auto& name_presence : src.flag_presents)
    {
        (*dst)[name_presence.first] = name_presence.second;
    }
}

void proto_wrap(
        const Domain::DomainStateHistory::Record& record,
        const std::vector<std::string>& flags_names,
        Api::Domain::DomainStateHistory::Record* dst)
{
    if (record.presents.size() != flags_names.size())
    {
        throw std::runtime_error("number of flags must be equal to the number of flags names");
    }
    proto_wrap(record.valid_from, dst->mutable_valid_from());
    for (std::size_t idx = 0; idx < record.presents.size(); ++idx)
    {
        (*(dst->mutable_flags()))[flags_names[idx]] = record.presents[idx];
    }
}

void proto_wrap(const Domain::DomainStateHistory& src, Api::Domain::DomainStateHistory* dst)
{
    dst->mutable_timeline()->Reserve(src.timeline.size());
    for (const auto& record : src.timeline)
    {
        proto_wrap(record, src.flags_names, dst->mutable_timeline()->Add());
    }
    if (src.valid_to == boost::none)
    {
        dst->clear_valid_to();
    }
    else
    {
        proto_wrap(*src.valid_to, dst->mutable_valid_to());
    }
}

void proto_wrap(const Domain::DomainDataHistory::Record& src, Api::Domain::DomainHistoryRecord* dst)
{
    proto_wrap(src.domain_history_id, dst->mutable_domain_history_id());
    proto_wrap(src.valid_from, dst->mutable_valid_from());
    proto_wrap(src.log_entry_id, dst->mutable_log_entry_id());
}

void proto_wrap(const std::vector<Domain::DomainDataHistory::Record>& src,
                decltype(std::declval<Api::Domain::DomainHistoryReply::Data>().mutable_timeline()) dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Domain::DomainDataHistory& src, Api::Domain::DomainHistoryReply::Data* dst)
{
    proto_wrap(src.domain_id, dst->mutable_domain_id());
    proto_wrap(src.timeline, dst->mutable_timeline());
    if (src.valid_to != boost::none)
    {
        proto_wrap(*src.valid_to, dst->mutable_valid_to());
    }
    else
    {
        dst->clear_valid_to();
    }
}

void proto_wrap(const Domain::DomainDataHistoryReply::Data& src, Api::Domain::DomainHistoryReply::Data* dst)
{
    proto_wrap(src.history, dst);
}

void proto_wrap(const Domain::DomainLifetime::TimeSpec& src, Api::Domain::DomainLifetime::TimeSpec* dst)
{
    proto_wrap(src.domain_history_id, dst->mutable_domain_history_id());
    proto_wrap(src.timestamp, dst->mutable_timestamp());
}

void proto_wrap(const Domain::DomainLifetime& src, Api::Domain::DomainLifetime* dst)
{
    proto_wrap(src.domain_id, dst->mutable_domain_id());
    proto_wrap(src.begin, dst->mutable_registered_from());
    if (src.end != boost::none)
    {
        proto_wrap(*src.end, dst->mutable_registered_to());
    }
    else
    {
        dst->clear_registered_to();
    }
}

template <typename T>
void proto_wrap_vector_of_domain_lifetime(const std::vector<Domain::DomainLifetime>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Domain::FqdnHistoryReply::Data& src, Api::Domain::FqdnHistoryReply::Data* dst)
{
    proto_wrap(src.fqdn, dst->mutable_fqdn());
    proto_wrap_vector_of_domain_lifetime(src.timeline, dst->mutable_timeline());
}

void proto_wrap(const Domain::DomainStateFlagsInfo& src, Api::Domain::DomainStateFlagsReply::Data* dst)
{
    proto_wrap_domain_state_flags(src, dst->mutable_traits());
}

template <typename T>
void proto_wrap_set_of_domain_items(const std::set<Domain::DomainItem>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto item : src)
    {
        *(dst->Add()) = Util::Into<std::string>::from(item);
    }
}

void proto_wrap(const Domain::SearchDomainReply::Data::Result& src, Api::Domain::SearchDomainReply::Data::Result* dst)
{
    proto_wrap(src.domain_id, dst->mutable_object_id());
    proto_wrap_set_of_domain_items(src.matched_items, dst->mutable_matched_items());
}

template <typename T>
void proto_wrap_vector_of_search_domain_result(const std::vector<Domain::SearchDomainReply::Data::Result>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Domain::SearchDomainReply::Data::FuzzyValue& src,
                Api::Domain::SearchDomainReply::Data::FuzzyValue* dst)
{
    dst->set_lower_estimate(src.lower_estimate);
    dst->set_upper_estimate(src.upper_estimate);
}

void proto_wrap(const Domain::SearchDomainReply::Data& src, Api::Domain::SearchDomainReply::Data* dst)
{
    proto_wrap_vector_of_search_domain_result(src.most_similar_domains, dst->mutable_results());
    proto_wrap(src.result_count, dst->mutable_estimated_total());
    proto_wrap_set_of_domain_items(src.searched_items, dst->mutable_searched_items());
}

template <typename T>
void proto_wrap_vector_of_domain_history_id(
        const std::vector<Domain::DomainHistoryId>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Domain::SearchDomainHistoryReply::Data::Result::HistoryPeriod& src,
                Api::Domain::SearchDomainHistoryReply::Data::Result::HistoryPeriod* dst)
{
    proto_wrap_set_of_domain_items(src.matched_items, dst->mutable_matched_items());
    proto_wrap(src.valid_from, dst->mutable_valid_from());
    if (src.valid_to == boost::none)
    {
        dst->clear_valid_to();
    }
    else
    {
        proto_wrap(*src.valid_to, dst->mutable_valid_to());
    }
    proto_wrap_vector_of_domain_history_id(src.domain_history_ids, dst->mutable_domain_history_ids());
}

template <typename T>
void proto_wrap_vector_of_search_domain_history_result_histories(
        const std::vector<Domain::SearchDomainHistoryReply::Data::Result::HistoryPeriod>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Domain::SearchDomainHistoryReply::Data::Result& src,
                Api::Domain::SearchDomainHistoryReply::Data::Result* dst)
{
    proto_wrap(src.object_id, dst->mutable_object_id());
    proto_wrap_vector_of_search_domain_history_result_histories(src.histories, dst->mutable_histories());
}

template <typename T>
void proto_wrap_vector_of_search_domain_history_result(const std::vector<Domain::SearchDomainHistoryReply::Data::Result>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Domain::SearchDomainHistoryReply::Data::FuzzyValue& src,
                Api::Domain::SearchDomainHistoryReply::Data::FuzzyValue* dst)
{
    dst->set_lower_estimate(src.lower_estimate);
    dst->set_upper_estimate(src.upper_estimate);
}

void proto_wrap(const Domain::SearchDomainHistoryReply::Data& src, Api::Domain::SearchDomainHistoryReply::Data* dst)
{
    proto_wrap_vector_of_search_domain_history_result(src.results, dst->mutable_results());
    proto_wrap(src.result_count, dst->mutable_estimated_total());
    proto_wrap_set_of_domain_items(src.searched_items, dst->mutable_searched_items());
}

void proto_wrap(const Registrar::RegistrarHandle& src, Api::Registrar::RegistrarHandle* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Registrar::PostalCode& src, Api::PostalCode* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Registrar::CountryCode& src, Api::CountryCode* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Registrar::PlaceAddress& src, Api::PlaceAddress* dst)
{
    for (const auto& street : src.street)
    {
        dst->add_street(street);
    }
    dst->set_city(src.city);
    dst->set_state_or_province(src.state_or_province);
    proto_wrap(src.postal_code, dst->mutable_postal_code());
    proto_wrap(src.country_code, dst->mutable_country_code());
}

void proto_wrap(const Registrar::PhoneNumber& src, Api::PhoneNumber* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Registrar::EmailAddress& src, Api::EmailAddress* dst)
{
    dst->set_value(get_raw_value_from(src));
}

template <typename T>
void proto_wrap_emails(const std::vector<Registrar::EmailAddress>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& email : src)
    {
        proto_wrap(email, dst->Add());
    }
}

void proto_wrap(const Registrar::Url& src, Api::Url* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Registrar::VatIdentificationNumber& src, Api::VatIdentificationNumber* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Registrar::CompanyRegistrationNumber& src, Api::CompanyRegistrationNumber* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Registrar::RegistrarCreditReply::Data& src, Api::Registrar::RegistrarCreditReply::Data* dst)
{
    proto_wrap(src.registrar_handle, dst->mutable_registrar_handle());
    for (const auto& zone_credit : src.credit_by_zone)
    {
        Api::Registrar::Credit credit;
        if (zone_credit.second != boost::none)
        {
            credit.set_decimal_value(get_raw_value_from(*zone_credit.second));
        }
        dst->mutable_credit_by_zone()->insert({zone_credit.first, credit});
    }
}

void proto_wrap(const Registrar::RegistrarInfoReply::Data& src, Api::Registrar::RegistrarInfoReply::Data* dst)
{
    proto_wrap(src.registrar_handle, dst->mutable_registrar_handle());
    dst->set_name(src.name);
    dst->set_organization(src.organization);
    if (src.place != boost::none)
    {
        proto_wrap(*src.place, dst->mutable_place());
    }
    if (src.telephone != boost::none)
    {
        proto_wrap(*src.telephone, dst->mutable_telephone());
    }
    if (src.fax != boost::none)
    {
        proto_wrap(*src.fax, dst->mutable_fax());
    }
    proto_wrap_emails(src.emails, dst->mutable_emails());
    if (src.url != boost::none)
    {
        proto_wrap(*src.url, dst->mutable_url());
    }
    dst->set_is_system_registrar(src.is_system_registrar);
    if (src.company_registration_number != boost::none)
    {
        proto_wrap(*src.company_registration_number, dst->mutable_company_registration_number());
    }
    if (src.vat_identification_number != boost::none)
    {
        proto_wrap(*src.vat_identification_number, dst->mutable_vat_identification_number());
    }
    dst->set_variable_symbol(src.variable_symbol);
    dst->set_payment_memo_regex(src.payment_memo_regex);
    dst->set_is_vat_payer(src.is_vat_payer);
}

void proto_wrap(const Registrar::RegistrarListReply::Data& src, Api::Registrar::RegistrarListReply::Data* dst)
{
    auto* const dst_handles = dst->mutable_registrar_handles();
    dst_handles->Reserve(src.registrar_handles.size());
    for (const auto& handle : src.registrar_handles)
    {
        proto_wrap(handle, dst_handles->Add());
    }
}

void proto_wrap(const Registrar::RegistrarZoneAccessReply::Data& src, Api::Registrar::RegistrarZoneAccessReply::Data* dst)
{
    proto_wrap(src.registrar_handle, dst->mutable_registrar_handle());
    for (const auto& zone_access : src.access_by_zone)
    {
        Api::Registrar::RegistrarZoneAccessReply::Data::Access access;
        access.set_has_access_now(zone_access.second.has_access_now);
        dst->mutable_access_by_zone()->insert({zone_access.first, access});
    }
}

void proto_wrap(const Domain::GetDomainsByContactReply::Data::Domain& src,
                Api::Domain::DomainsByContactReply::Data::Domain* dst)
{
    proto_wrap(src.domain_id, dst->mutable_domain_id());
    proto_wrap(src.domain_history_id, dst->mutable_domain_history_id());
    dst->set_is_holder(src.is_holder);
    dst->set_is_admin_contact(src.is_admin_contact);
    dst->set_is_keyset_tech_contact(src.is_keyset_tech_contact);
    dst->set_is_nsset_tech_contact(src.is_nsset_tech_contact);
}

template <typename T>
void proto_wrap_vector_of_get_domains_by_contact_reply_data_domains(const std::vector<Domain::GetDomainsByContactReply::Data::Domain>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Domain::GetDomainsNotifyInfoReply::Data::Info& src,
                Api::Domain::GetDomainsNotifyInfoReply::Data::Info* dst)
{
    proto_wrap(src.domain_id, dst->mutable_domain_id());
    dst->set_fqdn(src.fqdn);
    proto_wrap(src.registrant_id, dst->mutable_registrant_id());
    proto_wrap(src.registrant_handle, dst->mutable_registrant_handle());
    dst->set_registrant_name(src.registrant_name);
    dst->set_registrant_organization(src.registrant_organization);
    auto* const contacts = dst->mutable_contacts();
    contacts->Reserve(src.emails.size() + src.phones.size());
    std::for_each(begin(src.emails), end(src.emails), [&](auto&& email)
    {
        contacts->Add()->set_email(get_raw_value_from(email));
    });
    std::for_each(begin(src.phones), end(src.phones), [&](auto&& phone)
    {
        contacts->Add()->set_telephone(get_raw_value_from(phone));
    });
    auto* const additional_contacts = dst->mutable_additional_contacts();
    additional_contacts->Reserve(src.additional_emails.size() + src.additional_phones.size());
    std::for_each(begin(src.additional_emails), end(src.additional_emails), [&](auto&& email)
    {
        additional_contacts->Add()->set_email(get_raw_value_from(email));
    });
    std::for_each(begin(src.additional_phones), end(src.additional_phones), [&](auto&& phone)
    {
        additional_contacts->Add()->set_telephone(get_raw_value_from(phone));
    });
}

template <typename T>
void proto_wrap_vector_of_get_domains_notify_info_reply_data_info(const std::vector<Domain::GetDomainsNotifyInfoReply::Data::Info>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Domain::GetDomainsByHolderReply::Data::Domain& src,
                Api::Domain::DomainsByHolderReply::Data::Domain* dst)
{
    proto_wrap(src.domain_id, dst->mutable_domain_id());
    proto_wrap(src.domain_history_id, dst->mutable_domain_history_id());
    proto_wrap(src.fqdn, dst->mutable_fqdn());
}

template <typename T>
void proto_wrap_vector_of_get_domains_by_holder_reply_data_domains(const std::vector<Domain::GetDomainsByHolderReply::Data::Domain>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const PaginationReply& src, Api::PaginationReply* dst)
{
    dst->set_next_page_token(src.next_page_token);
    if (src.items_left != boost::none)
    {
        dst->set_items_left(*src.items_left);
    }
}

}//namespace Fred::Registry::{anonymous}

template<>
void proto_wrap(const Domain::DomainInfoReply::Data& src, Api::Domain::DomainInfoReply* dst)
{
    dst->clear_exception();
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Domain::DomainInfoReply::Exception::DomainDoesNotExist&, Api::Domain::DomainInfoReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_domain_does_not_exist();
}

template<>
void proto_wrap(const Domain::DomainIdReply::Data& src, Api::Domain::DomainIdReply* dst)
{
    dst->clear_exception();
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Domain::DomainIdReply::Exception::DomainDoesNotExist&, Api::Domain::DomainIdReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_domain_does_not_exist();
}

template<>
void proto_wrap(const Domain::DomainStateFlagsInfo& src, Api::Domain::DomainStateFlagsReply* dst)
{
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Domain::DomainDataHistoryReply::Data& src, Api::Domain::DomainHistoryReply* dst)
{
    dst->clear_data();
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Domain::DomainDataHistoryReply::Exception::DomainDoesNotExist&,
                Api::Domain::DomainHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_domain_does_not_exist();
}

template<>
void proto_wrap(const Domain::DomainDataHistoryReply::Exception::InvalidHistoryInterval&,
                Api::Domain::DomainHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_invalid_history_interval();
}

template<>
void proto_wrap(const Domain::FqdnHistoryReply::Data& src, Api::Domain::FqdnHistoryReply* dst)
{
    dst->clear_data();
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Domain::DomainStateReply::Data& src, Api::Domain::DomainStateReply* dst)
{
    proto_wrap(src.domain_id, dst->mutable_data()->mutable_domain_id());
    proto_wrap_domain_state(src.state, dst->mutable_data()->mutable_state()->mutable_flags());
}

template<>
void proto_wrap(const Domain::DomainStateReply::Exception::DomainDoesNotExist&, Api::Domain::DomainStateReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_domain_does_not_exist();
}

template<>
void proto_wrap(const Domain::DomainStateHistoryReply::Data& src, Api::Domain::DomainStateHistoryReply* dst)
{
    proto_wrap(src.domain_id, dst->mutable_data()->mutable_domain_id());
    proto_wrap(src.history, dst->mutable_data()->mutable_history());
}

template<>
void proto_wrap(const Domain::DomainStateHistoryReply::Exception::DomainDoesNotExist&,
                Api::Domain::DomainStateHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_domain_does_not_exist();
}

template<>
void proto_wrap(const Domain::DomainStateHistoryReply::Exception::InvalidHistoryInterval&,
                Api::Domain::DomainStateHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_invalid_history_interval();
}

template<>
void proto_wrap(const Domain::SearchDomainReply::Data& src, Api::Domain::SearchDomainReply* dst)
{
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Domain::SearchDomainHistoryReply::Data& src, Api::Domain::SearchDomainHistoryReply* dst)
{
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Domain::SearchDomainHistoryReply::Exception::ChronologyViolation&, Api::Domain::SearchDomainHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_chronology_violation();
}

template<>
void proto_wrap(const Domain::BatchDeleteDomainsReply::NotDeleted::Reason& src, Api::Domain::BatchDeleteDomainsReply::NotDeleted* dst)
{
    switch (src)
    {
        case Domain::BatchDeleteDomainsReply::NotDeleted::Reason::undefined:
            dst->set_reason(Api::Domain::BatchDeleteDomainsReply::NotDeleted::undefined);
            return;
        case Domain::BatchDeleteDomainsReply::NotDeleted::Reason::domain_blocked:
            dst->set_reason(Api::Domain::BatchDeleteDomainsReply::NotDeleted::domain_blocked);
            return;
        case Domain::BatchDeleteDomainsReply::NotDeleted::Reason::domain_does_not_exist:
            dst->set_reason(Api::Domain::BatchDeleteDomainsReply::NotDeleted::domain_does_not_exist);
            return;
        case Domain::BatchDeleteDomainsReply::NotDeleted::Reason::domain_history_id_mismatch:
            dst->set_reason(Api::Domain::BatchDeleteDomainsReply::NotDeleted::domain_history_id_mismatch);
            return;
        case Domain::BatchDeleteDomainsReply::NotDeleted::Reason::insufficient_permissions:
            dst->set_reason(Api::Domain::BatchDeleteDomainsReply::NotDeleted::insufficient_permissions);
            return;
    }
    throw std::runtime_error{"unexpected value " + std::to_string(static_cast<int>(src)) + " of Domain::BatchDeleteDomainsReply::NotDeleted::Reason"};
}

template<>
void proto_wrap(const Domain::BatchDeleteDomainsReply::NotDeleted& src, Api::Domain::BatchDeleteDomainsReply::NotDeleted* dst)
{
    proto_wrap(src.domain_id, dst->mutable_domain_id());
    proto_wrap(src.reason, dst);
}

std::size_t proto_wrap(
        const Domain::BatchDeleteDomainsReply::Data& src,
        std::size_t offset,
        Api::Domain::BatchDeleteDomainsReply::Data* dst)
{
    static constexpr auto max_number_of_domains = 10000;
    dst->Clear();
    if (src.not_deleted_domains.size() <= offset)
    {
        return 0;
    }
    const auto domains_to_send = src.not_deleted_domains.size() - offset;
    const auto chunk_size = domains_to_send < max_number_of_domains ? domains_to_send
                                                                    : max_number_of_domains;
    const auto next_offset = offset + chunk_size;
    auto* const domains = dst->mutable_not_deleted_domains();
    domains->Reserve(chunk_size);
    std::for_each(begin(src.not_deleted_domains) + offset, begin(src.not_deleted_domains) + next_offset, [&](auto&& record)
            {
                proto_wrap(record, domains->Add());
            });
    const bool last_chunk = src.not_deleted_domains.size() <= next_offset;
    if (last_chunk)
    {
        return 0;
    }
    return next_offset;
}

template<>
void proto_wrap(const Domain::ManageDomainStateFlagsReply::Data::DomainStateFlagRequestInfo& src, Api::Domain::ManageDomainStateFlagsReply::Data::DomainStateFlagRequestInfo* dst)
{
    proto_wrap(src.request_id, dst->mutable_request_id());
    proto_wrap(src.domain_id, dst->mutable_domain_id());
    dst->set_state_flag(src.state_flag);
}

std::size_t proto_wrap(
        const Domain::ManageDomainStateFlagsReply::Data& src,
        std::size_t offset,
        Api::Domain::ManageDomainStateFlagsReply::Data* dst)
{
    static constexpr auto max_number_of_requests = 10000;
    dst->Clear();
    if (src.state_flag_requests_created.size() <= offset)
    {
        return 0;
    }
    const auto requests_to_send = src.state_flag_requests_created.size() - offset;
    const auto chunk_size = requests_to_send < max_number_of_requests ? requests_to_send
                                                                      : max_number_of_requests;
    const auto next_offset = offset + chunk_size;
    auto* const requests = dst->mutable_state_flag_requests_created();
    requests->Reserve(chunk_size);
    std::for_each(begin(src.state_flag_requests_created) + offset, begin(src.state_flag_requests_created) + next_offset, [&](auto&& record)
            {
                proto_wrap(record, requests->Add());
            });
    const bool last_chunk = src.state_flag_requests_created.size() <= next_offset;
    if (last_chunk)
    {
        return 0;
    }
    return next_offset;
}

template<>
void proto_wrap(const Domain::ManageDomainStateFlagsReply::Exception::DomainDoesNotExist&,
                Api::Domain::ManageDomainStateFlagsReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_domain_does_not_exist();
}

template<>
void proto_wrap(const Domain::ManageDomainStateFlagsReply::Exception::DomainStateFlagRequestDoesNotExist&,
                Api::Domain::ManageDomainStateFlagsReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_request_does_not_exist();
}

template<>
void proto_wrap(const Domain::ManageDomainStateFlagsReply::Exception::DomainStateFlagDoesNotExist&,
                Api::Domain::ManageDomainStateFlagsReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_state_flag_does_not_exist();
}

template<>
void proto_wrap(const Domain::ManageDomainStateFlagsReply::Exception::DomainHistoryIdMismatch&,
                Api::Domain::ManageDomainStateFlagsReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_domain_history_id_mismatch();
}

template<>
void proto_wrap(const Domain::GetDomainsByContactReply::Data& src, Api::Domain::DomainsByContactReply* dst)
{
    proto_wrap_vector_of_get_domains_by_contact_reply_data_domains(src.domains, dst->mutable_data()->mutable_domains());
    if (src.pagination != boost::none)
    {
        proto_wrap(*src.pagination, dst->mutable_data()->mutable_pagination());
    }
}

std::size_t proto_wrap(
        const Domain::GetDomainsByContactReply::Data& src,
        std::size_t offset,
        Api::Domain::DomainsByContactReply::Data* dst)
{
    static constexpr auto max_number_of_domains = 10000;
    dst->Clear();
    const bool first_chunk = offset == 0;
    if (first_chunk && src.pagination != boost::none)
    {
        proto_wrap(*src.pagination, dst->mutable_pagination());
    }
    if (src.domains.size() <= offset)
    {
        return 0;
    }
    const auto domains_to_send = src.domains.size() - offset;
    const auto chunk_size = domains_to_send < max_number_of_domains ? domains_to_send
                                                                    : max_number_of_domains;
    const auto next_offset = offset + chunk_size;
    auto* const domains = dst->mutable_domains();
    domains->Reserve(chunk_size);
    std::for_each(begin(src.domains) + offset, begin(src.domains) + next_offset, [&](auto&& record)
            {
                proto_wrap(record, domains->Add());
            });
    const bool last_chunk = src.domains.size() <= next_offset;
    if (last_chunk)
    {
        return 0;
    }
    return next_offset;
}

template<>
void proto_wrap(const Domain::GetDomainsNotifyInfoReply::Data& src, Api::Domain::GetDomainsNotifyInfoReply* dst)
{
    proto_wrap_vector_of_get_domains_notify_info_reply_data_info(src.domains, dst->mutable_data()->mutable_domains());
    if (src.pagination != boost::none)
    {
        proto_wrap(*src.pagination, dst->mutable_data()->mutable_pagination());
    }
}

std::size_t proto_wrap(
        const Domain::GetDomainsNotifyInfoReply::Data& src,
        std::size_t offset,
        Api::Domain::GetDomainsNotifyInfoReply::Data* dst)
{
    static constexpr auto max_number_of_domains = 10000;
    dst->Clear();
    const bool first_chunk = offset == 0;
    if (first_chunk && src.pagination != boost::none)
    {
        proto_wrap(*src.pagination, dst->mutable_pagination());
    }
    if (src.domains.size() <= offset)
    {
        return 0;
    }
    const auto domains_to_send = src.domains.size() - offset;
    const auto chunk_size = domains_to_send < max_number_of_domains ? domains_to_send
                                                                    : max_number_of_domains;
    const auto next_offset = offset + chunk_size;
    auto* const domains = dst->mutable_domains();
    domains->Reserve(chunk_size);
    std::for_each(begin(src.domains) + offset, begin(src.domains) + next_offset, [&](auto&& record)
            {
                proto_wrap(record, domains->Add());
            });
    const bool last_chunk = src.domains.size() <= next_offset;
    if (last_chunk)
    {
        return 0;
    }
    return next_offset;
}

template<>
void proto_wrap(const Domain::GetDomainsByHolderReply::Data& src, Api::Domain::DomainsByHolderReply* dst)
{
    proto_wrap(src.domain_holder_id, dst->mutable_data()->mutable_domain_holder_id());
    proto_wrap(src.domain_holder_history_id, dst->mutable_data()->mutable_domain_holder_history_id());
    proto_wrap_vector_of_get_domains_by_holder_reply_data_domains(src.domains, dst->mutable_data()->mutable_domains());
}

std::size_t proto_wrap(
        const Domain::GetDomainsByHolderReply::Data& src,
        std::size_t offset,
        Api::Domain::DomainsByHolderReply::Data* dst)
{
    static constexpr auto max_number_of_domains = 10000;
    dst->Clear();
    const bool first_chunk = offset == 0;
    if (first_chunk)
    {
        proto_wrap(src.domain_holder_id, dst->mutable_domain_holder_id());
        proto_wrap(src.domain_holder_history_id, dst->mutable_domain_holder_history_id());
    }
    if (src.domains.size() <= offset)
    {
        return 0;
    }
    const auto domains_to_send = src.domains.size() - offset;
    const auto chunk_size = domains_to_send < max_number_of_domains ? domains_to_send
                                                                    : max_number_of_domains;
    const auto next_offset = offset + chunk_size;
    auto* const domains = dst->mutable_domains();
    domains->Reserve(chunk_size);
    std::for_each(begin(src.domains) + offset, begin(src.domains) + next_offset, [&](auto&& record)
            {
                proto_wrap(record, domains->Add());
            });
    const bool last_chunk = src.domains.size() <= next_offset;
    if (last_chunk)
    {
        return 0;
    }
    return next_offset;
}

template<>
void proto_wrap(const Domain::GetDomainsByContactReply::Exception::ContactDoesNotExist&,
                Api::Domain::DomainsByContactReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_contact_does_not_exist();
}

template<>
void proto_wrap(const Domain::GetDomainsByContactReply::Exception::InvalidOrderBy&,
                Api::Domain::DomainsByContactReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_invalid_order_by();
}

template<>
void proto_wrap(const Domain::GetDomainsByHolderReply::Exception::DomainHolderDoesNotExist&,
                Api::Domain::DomainsByHolderReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_domain_holder_does_not_exist();
}

template<>
void proto_wrap(const Domain::UpdateDomainsAdditionalNotifyInfoReply::Exception::DomainDoesNotExist&,
                Api::Domain::UpdateDomainsAdditionalNotifyInfoReply* dst)
{
    dst->mutable_exception()->mutable_domain_does_not_exist();
}

template<>
void proto_wrap(const Domain::UpdateDomainsAdditionalNotifyInfoReply::Exception::InvalidDomainContactInfo& src,
                Api::Domain::UpdateDomainsAdditionalNotifyInfoReply* dst)
{
    dst->mutable_exception()->mutable_invalid_domain_contact_info()->set_invalid_value(src.invalid_value);
}

namespace {

void proto_wrap(const Keyset::KeysetHandle& src, Api::Keyset::KeysetHandle* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Keyset::KeysetId& src, Api::Keyset::KeysetId* dst)
{
    proto_wrap_uuid(src, dst->mutable_uuid());
}

void proto_wrap(const Keyset::KeysetHistoryId& src, Api::Keyset::KeysetHistoryId* dst)
{
    proto_wrap_uuid(src, dst->mutable_uuid());
}

void proto_wrap(const Keyset::DnsKey& src, Api::Keyset::DnsKey* dst)
{
    dst->set_flags(src.flags);
    dst->set_protocol(src.protocol);
    dst->set_alg(src.alg);
    dst->set_key(src.key);
}

template <typename T>
void proto_wrap_vector_of_dnskeys(const std::vector<Keyset::DnsKey>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Keyset::AuthInfo& src, Registry::Api::AuthInfo* dst)
{
    dst->set_password(src.password);
}

void proto_wrap(const Keyset::KeysetInfoReply::Data& src, Api::Keyset::KeysetInfoReply::Data* dst)
{
    proto_wrap(src.keyset_handle, dst->mutable_keyset_handle());
    proto_wrap(src.keyset_id, dst->mutable_keyset_id());
    proto_wrap(src.keyset_history_id, dst->mutable_keyset_history_id());
    proto_wrap_vector_of_dnskeys(src.dns_keys, dst->mutable_dns_keys());
    proto_wrap_vector_of_contacts(src.technical_contacts, dst->mutable_technical_contacts());
    proto_wrap(src.auth_info, dst->mutable_auth_info());
    proto_wrap(src.representative_events, dst->mutable_events());
    Registry::proto_wrap(src.sponsoring_registrar, dst->mutable_sponsoring_registrar());
}

void proto_wrap(const Keyset::KeysetIdReply::Data& src, Api::Keyset::KeysetIdReply::Data* dst)
{
    proto_wrap(src.keyset_id, dst->mutable_keyset_id());
    proto_wrap(src.keyset_history_id, dst->mutable_keyset_history_id());
}

template <typename T>
void proto_wrap_keyset_state_flags(const Keyset::KeysetStateFlagsInfo& src, T* dst)
{
    for (const auto& flag : src)
    {
        Registry::Api::StateFlag::Traits traits;
        switch (flag.how_to_set)
        {
            case Keyset::StateFlagInfo::Manipulation::automatic:
                traits.set_how_to_set(Registry::Api::StateFlag::Manipulation::automatic);
                break;
            case Keyset::StateFlagInfo::Manipulation::manual:
                traits.set_how_to_set(Registry::Api::StateFlag::Manipulation::manual);
                break;
        }
        switch (flag.visibility)
        {
            case Keyset::StateFlagInfo::Visibility::external:
                traits.set_visibility(Registry::Api::StateFlag::Visibility::external);
                break;
            case Keyset::StateFlagInfo::Visibility::internal:
                traits.set_visibility(Registry::Api::StateFlag::Visibility::internal);
                break;
        }
        (*dst)[flag.name] = traits;
    }
}

template <typename T>
void proto_wrap_keyset_state(const Keyset::KeysetState& src, T* dst)
{
    for (const auto& name_presence : src.flag_presents)
    {
        (*dst)[name_presence.first] = name_presence.second;
    }
}

void proto_wrap(
        const Keyset::KeysetStateHistory::Record& record,
        const std::vector<std::string>& flags_names,
        Api::Keyset::KeysetStateHistory::Record* dst)
{
    if (record.presents.size() != flags_names.size())
    {
        throw std::runtime_error("number of flags must be equal to the number of flags names");
    }
    proto_wrap(record.valid_from, dst->mutable_valid_from());
    for (std::size_t idx = 0; idx < record.presents.size(); ++idx)
    {
        (*(dst->mutable_flags()))[flags_names[idx]] = record.presents[idx];
    }
}

void proto_wrap(const Keyset::KeysetStateHistory& src, Api::Keyset::KeysetStateHistory* dst)
{
    dst->mutable_timeline()->Reserve(src.timeline.size());
    for (const auto& record : src.timeline)
    {
        proto_wrap(record, src.flags_names, dst->mutable_timeline()->Add());
    }
    if (src.valid_to == boost::none)
    {
        dst->clear_valid_to();
    }
    else
    {
        proto_wrap(*src.valid_to, dst->mutable_valid_to());
    }
}

void proto_wrap(const Keyset::KeysetDataHistory::Record& src, Api::Keyset::KeysetHistoryRecord* dst)
{
    proto_wrap(src.keyset_history_id, dst->mutable_keyset_history_id());
    proto_wrap(src.valid_from, dst->mutable_valid_from());
    proto_wrap(src.log_entry_id, dst->mutable_log_entry_id());
}

template <typename T>
void proto_wrap_vector_of_keyset_data_history_record(const std::vector<Keyset::KeysetDataHistory::Record>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Keyset::KeysetDataHistory& src, Api::Keyset::KeysetHistoryReply::Data* dst)
{
    proto_wrap(src.keyset_id, dst->mutable_keyset_id());
    proto_wrap_vector_of_keyset_data_history_record(src.timeline, dst->mutable_timeline());
    if (src.valid_to != boost::none)
    {
        proto_wrap(*src.valid_to, dst->mutable_valid_to());
    }
    else
    {
        dst->clear_valid_to();
    }
}

void proto_wrap(const Keyset::KeysetDataHistoryReply::Data& src, Api::Keyset::KeysetHistoryReply::Data* dst)
{
    proto_wrap(src.history, dst);
}

void proto_wrap(const Keyset::KeysetLifetime::TimeSpec& src, Api::Keyset::KeysetLifetime::TimeSpec* dst)
{
    proto_wrap(src.keyset_history_id, dst->mutable_keyset_history_id());
    proto_wrap(src.timestamp, dst->mutable_timestamp());
}

void proto_wrap(const Keyset::KeysetLifetime& src, Api::Keyset::KeysetLifetime* dst)
{
    proto_wrap(src.keyset_id, dst->mutable_keyset_id());
    proto_wrap(src.begin, dst->mutable_registered_from());
    if (src.end != boost::none)
    {
        proto_wrap(*src.end, dst->mutable_registered_to());
    }
    else
    {
        dst->clear_registered_to();
    }
}

template <typename T>
void proto_wrap_vector_of_keyset_lifetime(const std::vector<Keyset::KeysetLifetime>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Keyset::KeysetHandleHistoryReply::Data& src, Api::Keyset::KeysetHandleHistoryReply::Data* dst)
{
    proto_wrap(src.keyset_handle, dst->mutable_keyset_handle());
    proto_wrap_vector_of_keyset_lifetime(src.timeline, dst->mutable_timeline());
}

void proto_wrap(const Keyset::KeysetStateFlagsInfo& src, Api::Keyset::KeysetStateFlagsReply::Data* dst)
{
    proto_wrap_keyset_state_flags(src, dst->mutable_traits());
}

void proto_wrap(Keyset::KeysetItem src, std::string& dst)
{
    switch (src)
    {
        case Keyset::KeysetItem::keyset_handle:
            dst = "keyset_handle";
            return;
        case Keyset::KeysetItem::public_key:
            dst = "dns_keys.key";
            return;
    }
    throw std::runtime_error("unexpected value of Keyset::KeysetItem");
}

template <typename T>
void proto_wrap_set_of_keyset_matched_items(const std::set<Keyset::KeysetItem>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto item : src)
    {
        proto_wrap(item, *(dst->Add()));
    }
}

void proto_wrap(const Keyset::SearchKeysetReply::Data::Result& src, Api::Keyset::SearchKeysetReply::Data::Result* dst)
{
    proto_wrap(src.keyset_id, dst->mutable_object_id());
    proto_wrap_set_of_keyset_matched_items(src.matched_items, dst->mutable_matched_items());
}

template <typename T>
void proto_wrap_vector_of_search_keyset_result(const std::vector<Keyset::SearchKeysetReply::Data::Result>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Keyset::SearchKeysetReply::Data::FuzzyValue& src,
                Api::Keyset::SearchKeysetReply::Data::FuzzyValue* dst)
{
    dst->set_lower_estimate(src.lower_estimate);
    dst->set_upper_estimate(src.upper_estimate);
}

template <typename T>
void proto_wrap_set_of_keyset_items(const std::set<Keyset::KeysetItem>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto item : src)
    {
        *(dst->Add()) = Util::Into<std::string>::from(item);
    }
}

void proto_wrap(const Keyset::SearchKeysetReply::Data& src, Api::Keyset::SearchKeysetReply::Data* dst)
{
    proto_wrap_vector_of_search_keyset_result(src.most_similar_keysets, dst->mutable_results());
    proto_wrap(src.result_count, dst->mutable_estimated_total());
    proto_wrap_set_of_keyset_items(src.searched_items, dst->mutable_searched_items());
}

template <typename T>
void proto_wrap_vector_of_keyset_history_id(
    const std::vector<Keyset::KeysetHistoryId>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Keyset::SearchKeysetHistoryReply::Data::Result::HistoryPeriod& src,
                Api::Keyset::SearchKeysetHistoryReply::Data::Result::HistoryPeriod* dst)
{
    proto_wrap_set_of_keyset_items(src.matched_items, dst->mutable_matched_items());
    proto_wrap(src.valid_from, dst->mutable_valid_from());
    if (src.valid_to == boost::none)
    {
        dst->clear_valid_to();
    }
    else
    {
        proto_wrap(*src.valid_to, dst->mutable_valid_to());
    }
    proto_wrap_vector_of_keyset_history_id(src.keyset_history_ids, dst->mutable_keyset_history_ids());
}

template <typename T>
void proto_wrap_vector_of_search_keyset_history_result_histories(
    const std::vector<Keyset::SearchKeysetHistoryReply::Data::Result::HistoryPeriod>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Keyset::SearchKeysetHistoryReply::Data::Result& src,
                Api::Keyset::SearchKeysetHistoryReply::Data::Result* dst)
{
    proto_wrap(src.object_id, dst->mutable_object_id());
    proto_wrap_vector_of_search_keyset_history_result_histories(src.histories, dst->mutable_histories());
}

template <typename T>
void proto_wrap_vector_of_search_keyset_history_result(const std::vector<Keyset::SearchKeysetHistoryReply::Data::Result>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Keyset::SearchKeysetHistoryReply::Data::FuzzyValue& src,
                Api::Keyset::SearchKeysetHistoryReply::Data::FuzzyValue* dst)
{
    dst->set_lower_estimate(src.lower_estimate);
    dst->set_upper_estimate(src.upper_estimate);
}

void proto_wrap(const Keyset::SearchKeysetHistoryReply::Data& src, Api::Keyset::SearchKeysetHistoryReply::Data* dst)
{
    proto_wrap_vector_of_search_keyset_history_result(src.results, dst->mutable_results());
    proto_wrap(src.result_count, dst->mutable_estimated_total());
    proto_wrap_set_of_keyset_items(src.searched_items, dst->mutable_searched_items());
}

}//namespace Fred::Registry::{anonymous}

template<>
void proto_wrap(const Keyset::KeysetInfoReply::Data& src, Api::Keyset::KeysetInfoReply* dst)
{
    dst->clear_exception();
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Keyset::KeysetInfoReply::Exception::KeysetDoesNotExist&, Api::Keyset::KeysetInfoReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_keyset_does_not_exist();
}

template<>
void proto_wrap(const Keyset::KeysetIdReply::Data& src, Api::Keyset::KeysetIdReply* dst)
{
    dst->clear_exception();
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Keyset::KeysetIdReply::Exception::KeysetDoesNotExist&, Api::Keyset::KeysetIdReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_keyset_does_not_exist();
}

template<>
void proto_wrap(const Keyset::KeysetStateFlagsInfo& src, Api::Keyset::KeysetStateFlagsReply* dst)
{
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Keyset::KeysetDataHistoryReply::Data& src, Api::Keyset::KeysetHistoryReply* dst)
{
    dst->clear_data();
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Keyset::KeysetDataHistoryReply::Exception::KeysetDoesNotExist&,
                Api::Keyset::KeysetHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_keyset_does_not_exist();
}

template<>
void proto_wrap(const Keyset::KeysetDataHistoryReply::Exception::InvalidHistoryInterval&,
                Api::Keyset::KeysetHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_invalid_history_interval();
}

template<>
void proto_wrap(const Keyset::KeysetHandleHistoryReply::Data& src, Api::Keyset::KeysetHandleHistoryReply* dst)
{
    dst->clear_data();
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Keyset::KeysetStateReply::Data& src, Api::Keyset::KeysetStateReply* dst)
{
    proto_wrap(src.keyset_id, dst->mutable_data()->mutable_keyset_id());
    proto_wrap_keyset_state(src.state, dst->mutable_data()->mutable_state()->mutable_flags());
}

template<>
void proto_wrap(const Keyset::KeysetStateReply::Exception::KeysetDoesNotExist&, Api::Keyset::KeysetStateReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_keyset_does_not_exist();
}

template<>
void proto_wrap(const Keyset::KeysetStateHistoryReply::Data& src, Api::Keyset::KeysetStateHistoryReply* dst)
{
    proto_wrap(src.keyset_id, dst->mutable_data()->mutable_keyset_id());
    proto_wrap(src.history, dst->mutable_data()->mutable_history());
}

template<>
void proto_wrap(const Keyset::KeysetStateHistoryReply::Exception::KeysetDoesNotExist&,
                Api::Keyset::KeysetStateHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_keyset_does_not_exist();
}

template<>
void proto_wrap(const Keyset::KeysetStateHistoryReply::Exception::InvalidHistoryInterval&,
                Api::Keyset::KeysetStateHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_invalid_history_interval();
}

template<>
void proto_wrap(const Keyset::SearchKeysetReply::Data& src, Api::Keyset::SearchKeysetReply* dst)
{
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Keyset::SearchKeysetHistoryReply::Data& src, Api::Keyset::SearchKeysetHistoryReply* dst)
{
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Keyset::SearchKeysetHistoryReply::Exception::ChronologyViolation&, Api::Keyset::SearchKeysetHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_chronology_violation();
}

namespace {

void proto_wrap(const Nsset::NssetHandle& src, Api::Nsset::NssetHandle* dst)
{
    dst->set_value(get_raw_value_from(src));
}

void proto_wrap(const Nsset::NssetId& src, Api::Nsset::NssetId* dst)
{
    proto_wrap_uuid(src, dst->mutable_uuid());
}

void proto_wrap(const Nsset::NssetHistoryId& src, Api::Nsset::NssetHistoryId* dst)
{
    proto_wrap_uuid(src, dst->mutable_uuid());
}

void proto_wrap(const Nsset::DnsHost& src, Api::Nsset::DnsHost* dst)
{
    dst->set_fqdn(src.fqdn);
    for (const auto& ip_address : src.ip_addresses)
    {
        dst->add_ip_addresses(ip_address.to_string());
    }
}

template <typename T>
void proto_wrap_vector_of_dnshosts(const std::vector<Nsset::DnsHost>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Nsset::TechnicalCheckLevel& src, Api::Nsset::TechnicalCheckLevel* dst)
{
    dst->set_value(src.technical_check_level);
}

void proto_wrap(const Nsset::AuthInfo& src, Registry::Api::AuthInfo* dst)
{
    dst->set_password(src.password);
}

void proto_wrap(const Nsset::NssetInfoReply::Data& src, Api::Nsset::NssetInfoReply::Data* dst)
{
    proto_wrap(src.nsset_handle, dst->mutable_nsset_handle());
    proto_wrap(src.nsset_id, dst->mutable_nsset_id());
    proto_wrap(src.nsset_history_id, dst->mutable_nsset_history_id());
    proto_wrap_vector_of_dnshosts(src.dns_hosts, dst->mutable_dns_hosts());
    proto_wrap_vector_of_contacts(src.technical_contacts, dst->mutable_technical_contacts());
    if (src.technical_check_level != boost::none)
    {
        proto_wrap(*src.technical_check_level, dst->mutable_technical_check_level());
    }
    else
    {
        dst->clear_technical_check_level();
    }
    proto_wrap(src.auth_info, dst->mutable_auth_info());
    proto_wrap(src.representative_events, dst->mutable_events());
    Registry::proto_wrap(src.sponsoring_registrar, dst->mutable_sponsoring_registrar());
}

void proto_wrap(const Nsset::NssetIdReply::Data& src, Api::Nsset::NssetIdReply::Data* dst)
{
    proto_wrap(src.nsset_id, dst->mutable_nsset_id());
    proto_wrap(src.nsset_history_id, dst->mutable_nsset_history_id());
}

template <typename T>
void proto_wrap_nsset_state_flags(const Nsset::NssetStateFlagsInfo& src, T* dst)
{
    for (const auto& flag : src)
    {
        Registry::Api::StateFlag::Traits traits;
        switch (flag.how_to_set)
        {
            case Nsset::StateFlagInfo::Manipulation::automatic:
                traits.set_how_to_set(Registry::Api::StateFlag::Manipulation::automatic);
                break;
            case Nsset::StateFlagInfo::Manipulation::manual:
                traits.set_how_to_set(Registry::Api::StateFlag::Manipulation::manual);
                break;
        }
        switch (flag.visibility)
        {
            case Nsset::StateFlagInfo::Visibility::external:
                traits.set_visibility(Registry::Api::StateFlag::Visibility::external);
                break;
            case Nsset::StateFlagInfo::Visibility::internal:
                traits.set_visibility(Registry::Api::StateFlag::Visibility::internal);
                break;
        }
        (*dst)[flag.name] = traits;
    }
}

template <typename T>
void proto_wrap_nsset_state(const Nsset::NssetState& src, T* dst)
{
    for (const auto& name_presence : src.flag_presents)
    {
        (*dst)[name_presence.first] = name_presence.second;
    }
}

void proto_wrap(
        const Nsset::NssetStateHistory::Record& record,
        const std::vector<std::string>& flags_names,
        Api::Nsset::NssetStateHistory::Record* dst)
{
    if (record.presents.size() != flags_names.size())
    {
        throw std::runtime_error("number of flags must be equal to the number of flags names");
    }
    proto_wrap(record.valid_from, dst->mutable_valid_from());
    for (std::size_t idx = 0; idx < record.presents.size(); ++idx)
    {
        (*(dst->mutable_flags()))[flags_names[idx]] = record.presents[idx];
    }
}

void proto_wrap(const Nsset::NssetStateHistory& src, Api::Nsset::NssetStateHistory* dst)
{
    dst->mutable_timeline()->Reserve(src.timeline.size());
    for (const auto& record : src.timeline)
    {
        proto_wrap(record, src.flags_names, dst->mutable_timeline()->Add());
    }
    if (src.valid_to == boost::none)
    {
        dst->clear_valid_to();
    }
    else
    {
        proto_wrap(*src.valid_to, dst->mutable_valid_to());
    }
}

void proto_wrap(const Nsset::NssetDataHistory::Record& src, Api::Nsset::NssetHistoryRecord* dst)
{
    proto_wrap(src.nsset_history_id, dst->mutable_nsset_history_id());
    proto_wrap(src.valid_from, dst->mutable_valid_from());
    proto_wrap(src.log_entry_id, dst->mutable_log_entry_id());
}

template <typename T>
void proto_wrap_vector_of_nsset_data_history_record(const std::vector<Nsset::NssetDataHistory::Record>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Nsset::NssetDataHistory& src, Api::Nsset::NssetHistoryReply::Data* dst)
{
    proto_wrap(src.nsset_id, dst->mutable_nsset_id());
    proto_wrap_vector_of_nsset_data_history_record(src.timeline, dst->mutable_timeline());
    if (src.valid_to != boost::none)
    {
        proto_wrap(*src.valid_to, dst->mutable_valid_to());
    }
    else
    {
        dst->clear_valid_to();
    }
}

void proto_wrap(const Nsset::NssetDataHistoryReply::Data& src, Api::Nsset::NssetHistoryReply::Data* dst)
{
    proto_wrap(src.history, dst);
}

void proto_wrap(const Nsset::NssetLifetime::TimeSpec& src, Api::Nsset::NssetLifetime::TimeSpec* dst)
{
    proto_wrap(src.nsset_history_id, dst->mutable_nsset_history_id());
    proto_wrap(src.timestamp, dst->mutable_timestamp());
}

void proto_wrap(const Nsset::NssetLifetime& src, Api::Nsset::NssetLifetime* dst)
{
    proto_wrap(src.nsset_id, dst->mutable_nsset_id());
    proto_wrap(src.begin, dst->mutable_registered_from());
    if (src.end != boost::none)
    {
        proto_wrap(*src.end, dst->mutable_registered_to());
    }
    else
    {
        dst->clear_registered_to();
    }
}

template <typename T>
void proto_wrap_vector_of_nsset_lifetime(const std::vector<Nsset::NssetLifetime>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Nsset::NssetHandleHistoryReply::Data& src, Api::Nsset::NssetHandleHistoryReply::Data* dst)
{
    proto_wrap(src.nsset_handle, dst->mutable_nsset_handle());
    proto_wrap_vector_of_nsset_lifetime(src.timeline, dst->mutable_timeline());
}

void proto_wrap(const Nsset::NssetStateFlagsInfo& src, Api::Nsset::NssetStateFlagsReply::Data* dst)
{
    proto_wrap_nsset_state_flags(src, dst->mutable_traits());
}

void proto_wrap(Nsset::NssetItem src, std::string& dst)
{
    switch (src)
    {
        case Nsset::NssetItem::nsset_handle:
            dst = "nsset_handle";
            return;
        case Nsset::NssetItem::dns_hosts_fqdn:
            dst = "dns_hosts.fqdn";
            return;
        case Nsset::NssetItem::dns_hosts_ip_addresses:
            dst = "dns_hosts.ip_addresses";
            return;
    }
    throw std::runtime_error("unexpected value of Nsset::NssetItem");
}

template <typename T>
void proto_wrap_set_of_nsset_matched_items(const std::set<Nsset::NssetItem>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto item : src)
    {
        proto_wrap(item, *(dst->Add()));
    }
}

void proto_wrap(const Nsset::SearchNssetReply::Data::Result& src, Api::Nsset::SearchNssetReply::Data::Result* dst)
{
    proto_wrap(src.nsset_id, dst->mutable_object_id());
    proto_wrap_set_of_nsset_matched_items(src.matched_items, dst->mutable_matched_items());
}

template <typename T>
void proto_wrap_vector_of_search_nsset_result(const std::vector<Nsset::SearchNssetReply::Data::Result>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Nsset::SearchNssetReply::Data::FuzzyValue& src,
                Api::Nsset::SearchNssetReply::Data::FuzzyValue* dst)
{
    dst->set_lower_estimate(src.lower_estimate);
    dst->set_upper_estimate(src.upper_estimate);
}

template <typename T>
void proto_wrap_set_of_nsset_items(const std::set<Nsset::NssetItem>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto item : src)
    {
        *(dst->Add()) = Util::Into<std::string>::from(item);
    }
}

void proto_wrap(const Nsset::SearchNssetReply::Data& src, Api::Nsset::SearchNssetReply::Data* dst)
{
    proto_wrap_vector_of_search_nsset_result(src.most_similar_nssets, dst->mutable_results());
    proto_wrap(src.result_count, dst->mutable_estimated_total());
    proto_wrap_set_of_nsset_items(src.searched_items, dst->mutable_searched_items());
}

template <typename T>
void proto_wrap_vector_of_nsset_history_id(
    const std::vector<Nsset::NssetHistoryId>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Nsset::SearchNssetHistoryReply::Data::Result::HistoryPeriod& src,
                Api::Nsset::SearchNssetHistoryReply::Data::Result::HistoryPeriod* dst)
{
    proto_wrap_set_of_nsset_items(src.matched_items, dst->mutable_matched_items());
    proto_wrap(src.valid_from, dst->mutable_valid_from());
    if (src.valid_to == boost::none)
    {
        dst->clear_valid_to();
    }
    else
    {
        proto_wrap(*src.valid_to, dst->mutable_valid_to());
    }
    proto_wrap_vector_of_nsset_history_id(src.nsset_history_ids, dst->mutable_nsset_history_ids());
}

template <typename T>
void proto_wrap_vector_of_search_nsset_history_result_histories(
    const std::vector<Nsset::SearchNssetHistoryReply::Data::Result::HistoryPeriod>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Nsset::SearchNssetHistoryReply::Data::Result& src,
                Api::Nsset::SearchNssetHistoryReply::Data::Result* dst)
{
    proto_wrap(src.object_id, dst->mutable_object_id());
    proto_wrap_vector_of_search_nsset_history_result_histories(src.histories, dst->mutable_histories());
}

template <typename T>
void proto_wrap_vector_of_search_nsset_history_result(const std::vector<Nsset::SearchNssetHistoryReply::Data::Result>& src, T* dst)
{
    dst->Reserve(src.size());
    for (const auto& record : src)
    {
        proto_wrap(record, dst->Add());
    }
}

void proto_wrap(const Nsset::SearchNssetHistoryReply::Data::FuzzyValue& src,
                Api::Nsset::SearchNssetHistoryReply::Data::FuzzyValue* dst)
{
    dst->set_lower_estimate(src.lower_estimate);
    dst->set_upper_estimate(src.upper_estimate);
}

void proto_wrap(const Nsset::SearchNssetHistoryReply::Data& src, Api::Nsset::SearchNssetHistoryReply::Data* dst)
{
    proto_wrap_vector_of_search_nsset_history_result(src.results, dst->mutable_results());
    proto_wrap(src.result_count, dst->mutable_estimated_total());
    proto_wrap_set_of_nsset_items(src.searched_items, dst->mutable_searched_items());
}

}//namespace Fred::Registry::{anonymous}

template<>
void proto_wrap(const Nsset::NssetInfoReply::Data& src, Api::Nsset::NssetInfoReply* dst)
{
    dst->clear_exception();
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Nsset::NssetInfoReply::Exception::NssetDoesNotExist&, Api::Nsset::NssetInfoReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_nsset_does_not_exist();
}

template<>
void proto_wrap(const Nsset::NssetIdReply::Data& src, Api::Nsset::NssetIdReply* dst)
{
    dst->clear_exception();
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Nsset::NssetIdReply::Exception::NssetDoesNotExist&, Api::Nsset::NssetIdReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_nsset_does_not_exist();
}

template<>
void proto_wrap(const Nsset::NssetStateFlagsInfo& src, Api::Nsset::NssetStateFlagsReply* dst)
{
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Nsset::NssetDataHistoryReply::Data& src, Api::Nsset::NssetHistoryReply* dst)
{
    dst->clear_data();
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Nsset::NssetDataHistoryReply::Exception::NssetDoesNotExist&,
                Api::Nsset::NssetHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_nsset_does_not_exist();
}

template<>
void proto_wrap(const Nsset::NssetDataHistoryReply::Exception::InvalidHistoryInterval&,
                Api::Nsset::NssetHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_invalid_history_interval();
}

template<>
void proto_wrap(const Nsset::NssetHandleHistoryReply::Data& src, Api::Nsset::NssetHandleHistoryReply* dst)
{
    dst->clear_data();
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Nsset::NssetStateReply::Data& src, Api::Nsset::NssetStateReply* dst)
{
    proto_wrap(src.nsset_id, dst->mutable_data()->mutable_nsset_id());
    proto_wrap_nsset_state(src.state, dst->mutable_data()->mutable_state()->mutable_flags());
}

template<>
void proto_wrap(const Nsset::NssetStateReply::Exception::NssetDoesNotExist&, Api::Nsset::NssetStateReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_nsset_does_not_exist();
}

template<>
void proto_wrap(const Nsset::NssetStateHistoryReply::Data& src, Api::Nsset::NssetStateHistoryReply* dst)
{
    proto_wrap(src.nsset_id, dst->mutable_data()->mutable_nsset_id());
    proto_wrap(src.history, dst->mutable_data()->mutable_history());
}

template<>
void proto_wrap(const Nsset::NssetStateHistoryReply::Exception::NssetDoesNotExist&,
                Api::Nsset::NssetStateHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_nsset_does_not_exist();
}

template<>
void proto_wrap(const Nsset::NssetStateHistoryReply::Exception::InvalidHistoryInterval&,
                Api::Nsset::NssetStateHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_invalid_history_interval();
}

template<>
void proto_wrap(const Nsset::SearchNssetReply::Data& src, Api::Nsset::SearchNssetReply* dst)
{
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Nsset::SearchNssetHistoryReply::Data& src, Api::Nsset::SearchNssetHistoryReply* dst)
{
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Nsset::SearchNssetHistoryReply::Exception::ChronologyViolation&, Api::Nsset::SearchNssetHistoryReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_chronology_violation();
}


template<>
void proto_wrap(const Registrar::RegistrarCreditReply::Data& src, Api::Registrar::RegistrarCreditReply* dst)
{
    dst->clear_exception();
    proto_wrap(src, dst->mutable_data());
}

template <>
void proto_wrap(const Registrar::RegistrarCreditReply::Exception::RegistrarDoesNotExist&,
                Api::Registrar::RegistrarCreditReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_registrar_does_not_exist();
}

void proto_wrap(const Nsset::SearchNssetReply::Data& src, Api::Nsset::SearchNssetReply* dst)
{
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Registrar::RegistrarInfoReply::Data& src, Api::Registrar::RegistrarInfoReply* dst)
{
    dst->clear_exception();
    proto_wrap(src, dst->mutable_data());
}

template <>
void proto_wrap(const Registrar::RegistrarInfoReply::Exception::RegistrarDoesNotExist&, Api::Registrar::RegistrarInfoReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_registrar_does_not_exist();
}

template<>
void proto_wrap(const Registrar::RegistrarListReply::Data& src, Api::Registrar::RegistrarListReply* dst)
{
    proto_wrap(src, dst->mutable_data());
}

template<>
void proto_wrap(const Registrar::RegistrarZoneAccessReply::Data& src, Api::Registrar::RegistrarZoneAccessReply* dst)
{
    dst->clear_exception();
    proto_wrap(src, dst->mutable_data());
}

template <>
void proto_wrap(const Registrar::RegistrarZoneAccessReply::Exception::RegistrarDoesNotExist&,
                Api::Registrar::RegistrarZoneAccessReply* dst)
{
    dst->clear_data();
    dst->mutable_exception()->mutable_registrar_does_not_exist();
}

}//namespace Fred::Registry
}//namespace Fred
