/*
 * Copyright (C) 2019-2020  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */
#include "src/registrar/registrar_zone_access.hh"

#include "src/util/strong_type.hh"
#include "src/util/struct_to_string.hh"
#include "src/fred_unwrap.hh"

#include "libfred/registrar/zone_access/get_zone_access_history.hh"
#include "libfred/registrar/zone_access/exceptions.hh"

#include <boost/date_time/gregorian/gregorian.hpp>

namespace Fred {
namespace Registry {
namespace Registrar {

std::string RegistrarZoneAccessRequest::to_string() const
{
    return Util::StructToString().add("registrar_handle", registrar_handle)
                                 .finish();
}

std::string RegistrarZoneAccessReply::Data::to_string() const
{
    return Util::StructToString().add("registrar_handle", registrar_handle)
                                 .add("access_by_zone", access_by_zone)
                                 .finish();
}

RegistrarZoneAccessReply::Exception::RegistrarDoesNotExist::RegistrarDoesNotExist(const RegistrarZoneAccessRequest& src)
    : Registrar::Exception("registrar " + src.to_string() + " does not exist")
{
}

RegistrarZoneAccessReply::Data registrar_zone_access(const RegistrarZoneAccessRequest& request)
{
    LibFred::OperationContextCreator ctx;
    try
    {
        const auto zone_access_history =
                LibFred::Registrar::ZoneAccess::GetZoneAccessHistory(get_raw_value_from(request.registrar_handle)).exec(ctx);
        const auto today = boost::gregorian::day_clock::local_day();
        RegistrarZoneAccessReply::Data result;
        for (const auto& zone_invoices : zone_access_history.invoices_by_zone)
        {
            const auto& zone_fqdn = zone_invoices.first;
            ZoneAccess zone_access;
            zone_access.has_access_now = has_access(zone_access_history, zone_fqdn, today);
            result.access_by_zone.insert({zone_fqdn, zone_access});
        }
        result.registrar_handle = request.registrar_handle;
        return result;
    }
    catch (const LibFred::Registrar::ZoneAccess::NonexistentRegistrar&)
    {
        throw RegistrarZoneAccessReply::Exception::RegistrarDoesNotExist(request);
    }
    catch (const std::exception&)
    {
        throw InternalServerError("expected unknown exception");
    }
    catch (...)
    {
        throw InternalServerError("unexpected unknown exception");
    }
}

}//namespace Fred::Registry::Registrar
}//namespace Fred::Registry
}//namespace Fred
